# -*- coding: utf-8 -*-

from qtpy.QtCharts import QtCharts

from prettyqt import charts


QtCharts.QStackedBarSeries.__bases__ = (charts.AbstractBarSeries,)


class StackedBarSeries(QtCharts.QStackedBarSeries):
    pass



# -*- coding: utf-8 -*-

from qtpy import QtNetwork

from prettyqt.utils import bidict

RAW_FORMS = bidict(
    name_and_value_only=QtNetwork.QNetworkCookie.NameAndValueOnly,
    full=QtNetwork.QNetworkCookie.Full,
)


class NetworkCookie(QtNetwork.QNetworkCookie):
    def __repr__(self):
        return "NetworkCookie()"

    def to_raw_form(self, full: bool = True):
        form = RAW_FORMS["full"] if full else RAW_FORMS["name_and_value_only"]
        self.toRawForm(form)

    def set_name(self, name: str):
        self.setName(str.encode(name))

    def get_name(self) -> str:
        return bytes(self.name()).decode()

    def set_value(self, value: str):
        self.setValue(str.encode(value))

    def get_value(self) -> str:
        return bytes(self.value()).decode()



# -*- coding: utf-8 -*-

from qtpy import QtGui

from prettyqt import gui
from prettyqt.utils import bidict

UPDATE_BEHAVIOUR = bidict(
    no_partial=QtGui.QOpenGLWindow.NoPartialUpdate,
    partial_blit=QtGui.QOpenGLWindow.PartialUpdateBlit,
    partial_blend=QtGui.QOpenGLWindow.PartialUpdateBlend,
)


QtGui.QOpenGLWindow.__bases__ = (gui.PaintDeviceWindow,)


class OpenGLWindow(QtGui.QOpenGLWindow):
    def __bool__(self):
        return self.isValid()

    def get_update_behaviour(self) -> str:
        """Get the window update hehaviour.

        Possible values: "no_partial", "partial_blit", "partial_blend"

        Returns:
            update behaviour
        """
        return UPDATE_BEHAVIOUR.inv[self.updateBehavior()]

    def grab_framebuffer(self) -> gui.Image:
        return gui.Image(self.grabFramebuffer())


if __name__ == "__main__":
    app = gui.GuiApplication([])
    wnd = OpenGLWindow()
    print(bool(wnd))
    wnd.show()
    print(bool(wnd))



# -*- coding: utf-8 -*-

"""Custom_models module.

Contains custom models
"""


from .selectionmixin import SelectionMixin
from .playlistmodel import PlaylistModel
from .transposeproxymodel import TransposeProxyModel
from .regexmatchesmodel import RegexMatchesModel
from .columnitemmodel import ColumnItemModel

__all__ = [
    "SelectionMixin",
    "TransposeProxyModel",
    "PlaylistModel",
    "RegexMatchesModel",
    "ColumnItemModel",
]



# -*- coding: utf-8 -*-

from typing import Tuple

from qtpy import QtCore, QtWidgets

from prettyqt import widgets


QtWidgets.QTableWidget.__bases__ = (widgets.TableView,)


class TableWidget(QtWidgets.QTableWidget):
    def sort(self, column: int = 0, reverse: bool = False):
        order = QtCore.Qt.DescendingOrder if reverse else QtCore.Qt.AscendingOrder
        self.sortItems(column, order)

    def __getitem__(self, index: Tuple[int, int]) -> QtWidgets.QTableWidgetItem:
        return self.item(*index)

    def __setitem__(self, index: Tuple[int, int], value: QtWidgets.QTableWidgetItem):
        self.setItem(index[0], index[1], value)

    def __delitem__(self, index: Tuple[int, int]):
        self.takeItem(*index)


if __name__ == "__main__":
    app = QtWidgets.QApplication([])
    widget = TableWidget()
    widget.setHorizontalHeaderLabels(["testus"])
    widget.setColumnCount(1)
    widget.insertRow(0)
    widget[0, 0] = widgets.TableWidgetItem("test")
    widget.show()
    app.main_loop()



"""Provides QtCore classes and functions."""

from prettyqt.qt import PYQT5, PYSIDE2, PythonQtError


if PYQT5:
    from PyQt5.QtCore import *
    from PyQt5.QtCore import pyqtSignal as Signal
    from PyQt5.QtCore import pyqtBoundSignal as SignalInstance
    from PyQt5.QtCore import pyqtSlot as Slot
    from PyQt5.QtCore import pyqtProperty as Property
    from PyQt5.QtCore import QT_VERSION_STR as __version__
    from PyQt5.QtCore import pyqtSignal
    from PyQt5.QtCore import pyqtBoundSignal
    from PyQt5.QtCore import pyqtSlot
    from PyQt5.QtCore import pyqtProperty
    from PyQt5.QtCore import QT_VERSION_STR

    # For issue #153
    from PyQt5.QtCore import QDateTime

    QDateTime.toPython = QDateTime.toPyDateTime

    # Those are imported from `import *`
    del pyqtSignal, pyqtBoundSignal, pyqtSlot, pyqtProperty, QT_VERSION_STR
elif PYSIDE2:
    from PySide2.QtCore import *
    from PySide2.QtCore import __version__



# -*- coding: utf-8 -*-

from typing import Optional
import pathlib

from qtpy import QtCore

from prettyqt import core


QtCore.QTranslator.__bases__ = (core.Object,)


class Translator(QtCore.QTranslator):
    def get_file_path(self) -> Optional[pathlib.Path]:
        path = self.filePath()
        if not path:
            return None
        return pathlib.Path(path)



# -*- coding: utf-8 -*-

from qtpy import PYQT5, PYSIDE2

if PYQT5:
    from PyQt5 import QtPositioning
elif PYSIDE2:
    from PySide2 import QtPositioning

from prettyqt import core, positioning


class GeoAreaMonitorInfo(QtPositioning.QGeoAreaMonitorInfo):
    def __str__(self):
        return self.name()

    def __repr__(self):
        return f"GeoAreaMonitorInfo({self.name()!r})"

    def get_area(self) -> positioning.GeoShape:
        area = self.area()
        if isinstance(area, QtPositioning.QGeoCircle):
            return positioning.GeoCircle(area)
        elif isinstance(area, QtPositioning.QGeoRectangle):
            return positioning.GeoRectangle(area)
        elif isinstance(area, QtPositioning.QGeoPath):
            return positioning.GeoPath(area)
        elif isinstance(area, QtPositioning.QGeoPolygon):
            return positioning.GeoPolygon(area)
        elif isinstance(area, QtPositioning.QGeoShape):
            return positioning.GeoShape(area)
        else:
            raise RuntimeError()

    def get_expiration(self) -> core.DateTime:
        return core.DateTime(self.expiration())


if __name__ == "__main__":
    info = GeoAreaMonitorInfo("test")
    print(str(info))
    print(repr(info))



# -*- coding: utf-8 -*-

from qtpy import QtCore


class Date(QtCore.QDate):
    def __repr__(self):
        template = super().__repr__().split("(")[1]
        return f"Date({template}"

    def __str__(self):
        return self.toString("yyyy-MM-dd")

    def __reduce__(self):
        return (self.__class__, (self.year(), self.month(), self.day()))


if __name__ == "__main__":
    dt = Date(2000, 11, 11)
    print(dt)



# -*- coding: utf-8 -*-

from typing import Callable

from qtpy import QtCore

from prettyqt.utils import bidict, InvalidParamError

TYPES = bidict(
    linear=QtCore.QEasingCurve.Linear,
    in_quad=QtCore.QEasingCurve.InQuad,
    out_quad=QtCore.QEasingCurve.OutQuad,
    in_out_quad=QtCore.QEasingCurve.InOutQuad,
    out_in_quad=QtCore.QEasingCurve.OutInQuad,
    in_cubic=QtCore.QEasingCurve.InCubic,
    out_cubic=QtCore.QEasingCurve.OutCubic,
    in_out_cubic=QtCore.QEasingCurve.InOutCubic,
    out_in_cubic=QtCore.QEasingCurve.OutInCubic,
    in_quart=QtCore.QEasingCurve.InQuart,
    out_quart=QtCore.QEasingCurve.OutQuart,
    in_out_quart=QtCore.QEasingCurve.InOutQuart,
    out_in_quart=QtCore.QEasingCurve.OutInQuart,
    in_quint=QtCore.QEasingCurve.InQuint,
    out_quint=QtCore.QEasingCurve.OutQuint,
    in_out_quint=QtCore.QEasingCurve.InOutQuint,
    out_in_quint=QtCore.QEasingCurve.OutInQuint,
    in_sine=QtCore.QEasingCurve.InSine,
    out_sine=QtCore.QEasingCurve.OutSine,
    in_out_sine=QtCore.QEasingCurve.InOutSine,
    out_in_sine=QtCore.QEasingCurve.OutInSine,
    in_expo=QtCore.QEasingCurve.InExpo,
    out_expo=QtCore.QEasingCurve.OutExpo,
    in_out_expo=QtCore.QEasingCurve.InOutExpo,
    out_in_expo=QtCore.QEasingCurve.OutInExpo,
    in_circ=QtCore.QEasingCurve.InCirc,
    out_circ=QtCore.QEasingCurve.OutCirc,
    in_out_circ=QtCore.QEasingCurve.InOutCirc,
    out_in_circ=QtCore.QEasingCurve.OutInCirc,
    in_elastic=QtCore.QEasingCurve.InElastic,
    out_elastic=QtCore.QEasingCurve.OutElastic,
    in_out_elastic=QtCore.QEasingCurve.InOutElastic,
    out_in_elastic=QtCore.QEasingCurve.OutInElastic,
    in_back=QtCore.QEasingCurve.InBack,
    out_back=QtCore.QEasingCurve.OutBack,
    in_out_back=QtCore.QEasingCurve.InOutBack,
    out_in_back=QtCore.QEasingCurve.OutInBack,
    in_bounce=QtCore.QEasingCurve.InBounce,
    out_bounce=QtCore.QEasingCurve.OutBounce,
    in_out_bounce=QtCore.QEasingCurve.InOutBounce,
    out_in_bounce=QtCore.QEasingCurve.OutInBounce,
    bezier_spline=QtCore.QEasingCurve.BezierSpline,
    tcb_spline=QtCore.QEasingCurve.TCBSpline,
    custom=QtCore.QEasingCurve.Custom,
)


class EasingCurve(QtCore.QEasingCurve):
    def __getitem__(self, value: float) -> float:
        return self.valueForProgress(value)

    def set_custom_type(self, method: Callable[[float], float]):
        self.setCustomType(method)

    def get_custom_type(self) -> Callable[[float], float]:
        return self.customType()

    def set_type(self, typ: str):
        """Set easing curve type.

        Valid values for modality: "linear", "in_quad", "out_quad", "in_out_quad",
        "out_in_quad", "in_cubic", "out_cubic", "in_out_cubic", "out_in_cubic",
        "in_quart", "out_quart", "in_out_quart", "out_in_quart", "in_quint",
        "out_quint", "in_out_quint", "out_in_quint", "in_sine", "out_sine",
        "in_out_sine", "out_in_sine", "in_expo", "out_expo", "in_out_expo",
        "out_in_expo", "in_circ", "out_circ", "in_out_circ", "out_in_circ",
        "in_elastic", "out_elastic", "in_out_elastic", "out_in_elastic",
        "in_back", "out_back", "in_out_back", "out_in_back", "in_bounce",
        "out_bounce", "in_out_bounce", "out_in_bounce", "bezier_spline",
        "tcb_spline", "custom"

        Args:
            typ: easing curve type

        Raises:
            InvalidParamError: easing curve type does not exist
        """
        if typ not in TYPES:
            raise InvalidParamError(typ, TYPES)
        self.setType(TYPES[typ])

    def get_type(self) -> str:
        """Get the current easing curve type.

        Possible values: "linear", "in_quad", "out_quad", "in_out_quad",
        "out_in_quad", "in_cubic", "out_cubic", "in_out_cubic", "out_in_cubic",
        "in_quart", "out_quart", "in_out_quart", "out_in_quart", "in_quint",
        "out_quint", "in_out_quint", "out_in_quint", "in_sine", "out_sine",
        "in_out_sine", "out_in_sine", "in_expo", "out_expo", "in_out_expo",
        "out_in_expo", "in_circ", "out_circ", "in_out_circ", "out_in_circ",
        "in_elastic", "out_elastic", "in_out_elastic", "out_in_elastic",
        "in_back", "out_back", "in_out_back", "out_in_back", "in_bounce",
        "out_bounce", "in_out_bounce", "out_in_bounce", "bezier_spline",
        "tcb_spline", "custom"

        Returns:
            easing curve type
        """
        return TYPES.inv[self.type()]


if __name__ == "__main__":
    c = EasingCurve()



# -*- coding: utf-8 -*-

from typing import Union

from qtpy.QtCharts import QtCharts

from prettyqt import core, gui
from prettyqt.utils import bidict


VALUE_POSITIONS = bidict(
    lower_extreme=QtCharts.QBoxSet.LowerExtreme,
    lower_quartile=QtCharts.QBoxSet.LowerQuartile,
    median=QtCharts.QBoxSet.Median,
    upper_quartile=QtCharts.QBoxSet.UpperQuartile,
    upper_extreme=QtCharts.QBoxSet.UpperExtreme,
)


QtCharts.QBoxSet.__bases__ = (core.Object,)


class BoxSet(QtCharts.QBoxSet):
    def __repr__(self):
        return (
            f"BoxSet({self['lower_extreme']}, {self['lower_quartile']}, {self['median']},"
            f" {self['upper_quartile']}, {self['upper_extreme']}, {self.label()!r})"
        )

    def __getitem__(self, index: Union[int, str]) -> float:
        if index in VALUE_POSITIONS:
            index = VALUE_POSITIONS[index]
        return self.at(index)

    def __setitem__(self, index: Union[int, str], value: int):
        if index in VALUE_POSITIONS:
            index = VALUE_POSITIONS[index]
        self.setValue(index, value)

    def get_pen(self) -> gui.Pen:
        return gui.Pen(self.pen())

    def get_brush(self) -> gui.Brush:
        return gui.Brush(self.brush())


if __name__ == "__main__":
    boxset = BoxSet()
    print(repr(boxset))



from qtpy import QtHelp

from prettyqt import qthelp


QtHelp.QHelpEngine.__bases__ = (qthelp.HelpEngineCore,)


class HelpEngine(QtHelp.QHelpEngine):
    pass


if __name__ == "__main__":
    engine = HelpEngine("")
    engine.get_files("a", "b")



# -*- coding: utf-8 -*-

from qtpy import QtCore

from prettyqt import core
from prettyqt.utils import bidict, InvalidParamError

DELETION_POLICIES = bidict(
    keep=QtCore.QAbstractAnimation.KeepWhenStopped,
    delete=QtCore.QAbstractAnimation.DeleteWhenStopped,
)

DIRECTIONS = bidict(
    forward=QtCore.QAbstractAnimation.Forward, backward=QtCore.QAbstractAnimation.Backward
)

STATES = bidict(
    stopped=QtCore.QAbstractAnimation.Stopped,
    paused=QtCore.QAbstractAnimation.Paused,
    running=QtCore.QAbstractAnimation.Running,
)

QtCore.QAbstractAnimation.__bases__ = (core.Object,)


class AbstractAnimation(QtCore.QAbstractAnimation):
    def set_direction(self, direction: str):
        """Set animation direction.

        Valid values: "forward", "backward"

        Args:
            direction: animation direction

        Raises:
            InvalidParamError: animation direction does not exist
        """
        if direction not in DIRECTIONS:
            raise InvalidParamError(direction, DIRECTIONS)
        self.setDirection(DIRECTIONS[direction])

    def get_direction(self) -> str:
        """Get the current animation direction.

        Possible values: "forward", "backward"

        Returns:
            animation direction
        """
        return DIRECTIONS.inv[self.direction()]

    def get_state(self) -> str:
        """Get the current animation state.

        Possible values: "stopped", "paused", "running"

        Returns:
            animation state
        """
        return STATES.inv[self.state()]

    def start_animation(self, policy: str):
        """Start the animation.

        Valid values for policy: "keep", "delete"

        Args:
            policy: animation policy

        Raises:
            InvalidParamError: animation policy does not exist
        """
        if policy not in DELETION_POLICIES:
            raise InvalidParamError(policy, DELETION_POLICIES)
        self.start(DELETION_POLICIES[policy])



# -*- coding: utf-8 -*-

from qtpy import QtCore

from prettyqt import core
from prettyqt.utils import bidict, InvalidParamError


QtCore.QTimer.__bases__ = (core.Object,)

TYPES = bidict(
    precise=QtCore.Qt.PreciseTimer,
    coarse=QtCore.Qt.CoarseTimer,
    very_coarse=QtCore.Qt.VeryCoarseTimer,
)


class Timer(QtCore.QTimer):
    @classmethod
    def single_shot(cls, callback):
        timer = cls()
        timer.timeout.connect(callback)
        timer.setSingleShot(True)
        return timer

    def set_type(self, typ: str):
        """Set the timer type.

        Allowed values are "precise", "coarse", "very_coarse"

        Args:
            typ: timer type

        Raises:
            InvalidParamError: timer type does not exist
        """
        if typ not in TYPES:
            raise InvalidParamError(typ, TYPES)
        self.setTimerType(TYPES[typ])

    def get_type(self) -> str:
        """Return current timer type.

        Possible values: "precise", "coarse", "very_coarse"

        Returns:
            timer type
        """
        return TYPES.inv[self.timerType()]



# -*- coding: utf-8 -*-

"""Webenginewidgets module.

contains QtWebEngineCore-based classes
"""

import logging

logger = logging.getLogger(__name__)

try:
    from .webenginehttprequest import WebEngineHttpRequest
    from .webengineurlscheme import WebEngineUrlScheme
    from .webengineurlschemehandler import WebEngineUrlSchemeHandler

    __all__ = ["WebEngineHttpRequest", "WebEngineUrlScheme", "WebEngineUrlSchemeHandler"]
except ModuleNotFoundError:
    logger.warning("Tried to import WebEngine module but not installed.")



# -*- coding: utf-8 -*-

from typing import Union, Dict

from qtpy import QtNetwork, QtCore

from prettyqt import core
from prettyqt.utils import bidict, InvalidParamError

KNOWN_HEADERS = bidict(
    content_disposition=QtNetwork.QNetworkRequest.ContentDispositionHeader,
    content_type=QtNetwork.QNetworkRequest.ContentTypeHeader,
    content_length=QtNetwork.QNetworkRequest.ContentLengthHeader,
    location=QtNetwork.QNetworkRequest.LocationHeader,
    last_modified=QtNetwork.QNetworkRequest.LastModifiedHeader,
    if_modified_since=QtNetwork.QNetworkRequest.IfModifiedSinceHeader,
    etag=QtNetwork.QNetworkRequest.ETagHeader,
    if_match=QtNetwork.QNetworkRequest.IfMatchHeader,
    if_none_match=QtNetwork.QNetworkRequest.IfNoneMatchHeader,
    cookie=QtNetwork.QNetworkRequest.CookieHeader,
    set_cooke=QtNetwork.QNetworkRequest.SetCookieHeader,
    user_agent=QtNetwork.QNetworkRequest.UserAgentHeader,
    server=QtNetwork.QNetworkRequest.ServerHeader,
)

PRIORITIES = bidict(
    high=QtNetwork.QNetworkRequest.HighPriority,
    normal=QtNetwork.QNetworkRequest.NormalPriority,
    low=QtNetwork.QNetworkRequest.LowPriority,
)

REDIRECT_POLICIES = bidict(
    manual=QtNetwork.QNetworkRequest.ManualRedirectPolicy,
    no_less_safe=QtNetwork.QNetworkRequest.NoLessSafeRedirectPolicy,
    same_origin=QtNetwork.QNetworkRequest.SameOriginRedirectPolicy,
    user_verified=QtNetwork.QNetworkRequest.UserVerifiedRedirectPolicy,
)

CACHE_LOAD_CONTROL = bidict(
    always_network=QtNetwork.QNetworkRequest.AlwaysNetwork,
    prefer_network=QtNetwork.QNetworkRequest.PreferNetwork,
    prefer_cache=QtNetwork.QNetworkRequest.PreferCache,
    always_cache=QtNetwork.QNetworkRequest.AlwaysCache,
)


class NetworkRequest(QtNetwork.QNetworkRequest):
    def __init__(self, obj: Union[QtCore.QUrl, str, QtNetwork.QNetworkRequest] = None):
        if isinstance(obj, QtNetwork.QNetworkRequest):
            super().__init__(obj)
        else:
            super().__init__()
            if obj is not None:
                self.set_url(obj)

    def __repr__(self):
        return f"NetworkRequest({self.get_url()})"

    def set_header(self, name: str, value: str):
        if name not in KNOWN_HEADERS:
            raise InvalidParamError(name, KNOWN_HEADERS)
        self.setHeader(KNOWN_HEADERS[name], value)

    def get_header(self, name: str) -> str:
        if name not in KNOWN_HEADERS:
            raise InvalidParamError(name, KNOWN_HEADERS)
        return self.header(KNOWN_HEADERS[name])

    def set_headers(self, headers: Dict[str, str]):
        for k, v in headers.items():
            self.setRawHeader(str.encode(k), str.encode(v))

    def get_headers(self) -> Dict[str, str]:
        return {
            bytes(h).decode(): bytes(self.rawHeader(h)).decode()
            for h in self.rawHeaderList()
        }

    def set_url(self, url: Union[str, QtCore.QUrl]):
        url = core.Url(url)
        self.setUrl(url)

    def get_url(self) -> core.Url:
        return core.Url(self.url())

    def set_priority(self, priority: str):
        """Set priority.

        Valid values for priority: "high", "normal", "low"

        Args:
            priority: priority

        Raises:
            InvalidParamError: priority does not exist
        """
        if priority not in PRIORITIES:
            raise InvalidParamError(priority, PRIORITIES)
        self.setPriority(PRIORITIES[priority])

    def get_priority(self) -> str:
        """Get the current priority.

        Possible values: "high", "normal", "low"

        Returns:
            priority
        """
        return PRIORITIES.inv[self.priority()]



# -*- coding: utf-8 -*-

from typing import Union

from qtpy import QtGui, QtWidgets

from prettyqt import gui, widgets
from prettyqt.utils import bidict, colors, InvalidParamError


VIEW_MODES = bidict(
    default=QtWidgets.QMdiArea.SubWindowView, tabbed=QtWidgets.QMdiArea.TabbedView
)


WINDOW_ORDERS = bidict(
    creation=QtWidgets.QMdiArea.CreationOrder,
    stacking=QtWidgets.QMdiArea.StackingOrder,
    activation_history=QtWidgets.QMdiArea.ActivationHistoryOrder,
)

TAB_POSITIONS = bidict(
    north=QtWidgets.QTabWidget.North,
    south=QtWidgets.QTabWidget.South,
    west=QtWidgets.QTabWidget.West,
    east=QtWidgets.QTabWidget.East,
)

QtWidgets.QMdiArea.__bases__ = (widgets.AbstractScrollArea,)


class MdiArea(QtWidgets.QMdiArea):
    def __add__(self, other):
        if isinstance(other, QtWidgets.QWidget):
            self.add(other)
            return self

    def serialize_fields(self):
        return dict(
            view_mode=self.get_view_mode(),
            window_order=self.get_window_order(),
            tab_position=self.get_tab_position(),
            background=self.get_background(),
            document_mode=self.documentMode(),
        )

    def set_view_mode(self, mode: str):
        """Set view mode for the MDI area.

        Valid values are "default", "tabbed"

        Args:
            mode: view mode to use

        Raises:
            InvalidParamError: view mode does not exist
        """
        if mode not in VIEW_MODES:
            raise InvalidParamError(mode, VIEW_MODES)
        self.setViewMode(VIEW_MODES[mode])

    def get_view_mode(self) -> str:
        """Return current view mode.

        Possible values: "default", "tabbed"

        Returns:
            view mode
        """
        return VIEW_MODES.inv[self.viewMode()]

    def set_window_order(self, mode: str):
        """Set the window order behaviour for the MDI area.

        Valid values are "creation", "stacking", "activation_history"

        Args:
            mode: window order behaviour to use

        Raises:
            InvalidParamError: window order mode not existing.
        """
        if mode not in WINDOW_ORDERS:
            raise InvalidParamError(mode, WINDOW_ORDERS)
        self.setActivationOrder(WINDOW_ORDERS[mode])

    def get_window_order(self) -> str:
        """Return current window order.

        Possible values: "creation", "stacking", "activation_history"

        Returns:
            view mode
        """
        return WINDOW_ORDERS.inv[self.activationOrder()]

    def set_tab_position(self, position: str):
        """Set tab position for the MDI area.

        Valid values are "north", "south", "west", "east"

        Args:
            position: tabs position to use

        Raises:
            InvalidParamError: tab position does not exist
        """
        if position not in TAB_POSITIONS:
            raise InvalidParamError(position, TAB_POSITIONS)
        self.setTabPosition(TAB_POSITIONS[position])

    def get_tab_position(self) -> str:
        """Return current tab position.

        Possible values: "north", "south", "west", "east"

        Returns:
            tab position
        """
        return TAB_POSITIONS.inv[self.tabPosition()]

    def set_background(
        self,
        brush_or_color: Union[QtGui.QBrush, colors.ColorType],
        pattern: str = "solid",
    ):
        if isinstance(brush_or_color, QtGui.QBrush):
            self.setBackground(brush_or_color)
        else:
            pattern = gui.painter.PATTERNS[pattern]
            color = colors.get_color(brush_or_color)
            brush = gui.Brush(color, pattern)
            self.setBackground(brush)

    def get_background(self) -> gui.Brush:
        return gui.Brush(self.background())

    def add(self, *item: QtWidgets.QWidget):
        for i in item:
            if not isinstance(i, QtWidgets.QMdiSubWindow):
                widget = widgets.MdiSubWindow()
                widget.setWidget(i)
                self.addSubWindow(widget)
            else:
                self.addSubWindow(i)


if __name__ == "__main__":
    app = widgets.app()
    widget = MdiArea()
    widget.set_background("green", "cross")
    le = widgets.LineEdit("test")
    le2 = widgets.LineEdit("test")
    widget.add(le)
    widget.add(le2)
    widget.show()
    app.main_loop()



# -*- coding: utf-8 -*-

from qtpy import QtCore

from prettyqt import core

QtCore.QSequentialAnimationGroup.__bases__ = (core.AnimationGroup,)


class SequentialAnimationGroup(QtCore.QSequentialAnimationGroup):
    pass



# -*- coding: utf-8 -*-
"""Syntax highlighter for the programming language Python."""

from prettyqt import core, gui, syntaxhighlighters


KEYWORDS = [
    "and",
    "assert",
    "break",
    "continue",
    "class",
    "def" "del",
    "elif",
    "else",
    "except",
    "exec",
    "finally",
    "for",
    "from",
    "global",
    "if",
    "import",
    "in",
    "is",
    "lambda",
    "not",
    "or",
    "pass",
    "print",
    "raise",
    "return",
    "try",
    "while",
    "yield",
    "None",
    "True",
    "False",
]

OPERATORS = [
    "=",
    "==",
    "!=",
    "<",
    "<=",
    ">",
    ">=",
    r"\+",
    "-",
    r"\*",
    "/",
    "//",
    r"\%",
    r"\*\*",
    r"\+=",
    "-=",
    r"\*=",
    "/=",
    r"\%=",
    r"\^",
    r"\|",
    r"\&",
    r"\~",
    ">>",
    "<<",
]


class Rule(syntaxhighlighters.HighlightRule):
    pass


class Keyword(Rule):
    regex = r"\b%s\b" % r"\b|\b".join(KEYWORDS)
    color = gui.Color(200, 120, 50)
    bold = True


class Operator(Rule):
    regex = r"%s" % "|".join(OPERATORS)
    color = gui.Color(150, 150, 150)


class Bracket(Rule):
    regex = r"[\[\]\{\}\(\)]"
    color = "darkgray"


class Self(Rule):
    regex = r"\bself\b"
    color = gui.Color(150, 85, 140)
    italic = True


class String(Rule):
    regex = [r'"[^"\\]*(\\.[^"\\]*)*"', r"'[^'\\]*(\\.[^'\\]*)*'"]
    color = gui.Color(20, 110, 100)


class Def(Rule):
    regex = r"\bdef\b\s*(\w+)"
    color = gui.Color(20, 20, 255)
    bold = True
    nth = 1


class Class(Rule):
    regex = r"\bclass\b\s*(\w+)"
    color = gui.Color(20, 20, 255)
    bold = True
    nth = 1


class Comment(Rule):
    regex = r"#[^\n]*"
    color = "lightgray"


class Number(Rule):
    regex = [
        r"\b[+-]?[0-9]+[lL]?\b",
        r"\b[+-]?0[xX][0-9A-Fa-f]+[lL]?\b",
        r"\b[+-]?[0-9]+(?:\.[0-9]+)?(?:[eE][+-]?[0-9]+)?\b",
    ]
    color = gui.Color(100, 150, 190)


# Multi-line strings (expression, flag, style)
# FIXME: The triple-quotes in these two lines will mess up the
# syntax highlighting from this point onward
fmt = gui.TextCharFormat()
fmt.set_foreground_color([30, 120, 110])
TRI_SINGLE = (core.RegExp("'''"), 1, fmt)
TRI_DOUBLE = (core.RegExp('"""'), 2, fmt)


class PythonHighlighter(gui.SyntaxHighlighter):
    """Syntax highlighter for the Python language."""

    RULES = Rule.__subclasses__()

    def highlightBlock(self, text: str):
        """Apply syntax highlighting to the given block of text."""
        # Do other syntax formatting
        super().highlightBlock(text)
        self.setCurrentBlockState(0)
        # Do multi-line strings
        if not self.match_multiline(text, *TRI_SINGLE):
            self.match_multiline(text, *TRI_DOUBLE)

    def match_multiline(self, text, delimiter, in_state, style):
        """Do highlighting of multi-line strings.

        ``delimiter`` should be a
        ``core.RegExp`` for triple-single-quotes or triple-double-quotes, and
        ``in_state`` should be a unique integer to represent the corresponding
        state changes when inside those strings. Returns True if we're still
        inside a multi-line string when this function is finished.
        """
        # If inside triple-single quotes, start at 0
        if self.previousBlockState() == in_state:
            start = 0
            add = 0
        # Otherwise, look for the delimiter on this line
        else:
            start = delimiter.indexIn(text)
            # Move past this match
            add = delimiter.matchedLength()

        # As long as there's a delimiter match on this line...
        while start >= 0:
            # Look for the ending delimiter
            end = delimiter.indexIn(text, start + add)
            # Ending delimiter on this line?
            if end >= add:
                length = end - start + add + delimiter.matchedLength()
                self.setCurrentBlockState(0)
            # No; multi-line string
            else:
                self.setCurrentBlockState(in_state)
                length = len(text) - start + add
            # Apply formatting
            self.setFormat(start, length, style)
            # Look for the next match
            start = delimiter.indexIn(text, start + length)

        # Return True if still inside a multi-line string, False otherwise
        return self.currentBlockState() == in_state


if __name__ == "__main__":
    from prettyqt import widgets

    app = widgets.app()
    editor = widgets.PlainTextEdit()
    highlighter = PythonHighlighter(editor.document())
    editor.show()
    app.main_loop()



# -*- coding: utf-8 -*-

from qtpy import QtLocation

from prettyqt import location
from prettyqt.utils import bidict

TYPE = bidict(
    none=QtLocation.QPlaceContent.NoType,
    image=QtLocation.QPlaceContent.ImageType,
    review=QtLocation.QPlaceContent.ReviewType,
    editorial=QtLocation.QPlaceContent.EditorialType,
    custom=QtLocation.QPlaceContent.CustomType,
)


class PlaceContent(QtLocation.QPlaceContent):
    def get_type(self) -> str:
        """Return the visibility of the place.

        Possible values are "none", "image", "review", "editorial", "custom"

        Returns:
            Place type
        """
        return TYPE.inv[self.type()]

    def get_user(self) -> location.PlaceUser:
        return location.PlaceUser(self.user())

    def get_supplier(self) -> location.PlaceSupplier:
        return location.PlaceSupplier(self.supplier())


if __name__ == "__main__":
    content = PlaceContent()



# -*- coding: utf-8 -*-

# from qtpy import QtWebEngineWidgets

from typing import Iterator, List

from qtpy import PYQT5, PYSIDE2

if PYQT5:
    from PyQt5 import QtWebEngineWidgets  # type: ignore
elif PYSIDE2:
    from PySide2 import QtWebEngineWidgets

from prettyqt import core, webenginewidgets


class WebEngineHistory(object):
    def __init__(self, history: QtWebEngineWidgets.QWebEngineHistory):
        self.history = history

    def __getattr__(self, val):
        return getattr(self.history, val)

    def __len__(self):
        return len(self.history)

    def __getitem__(self, index: int) -> webenginewidgets.WebEngineHistoryItem:
        item = self.history.itemAt(index)
        return webenginewidgets.WebEngineHistoryItem(item)

    def __iter__(self) -> Iterator[webenginewidgets.WebEngineHistoryItem]:
        items = [webenginewidgets.WebEngineHistoryItem(i) for i in self.history.items()]
        return iter(items)

    def __getstate__(self):
        return core.DataStream.create_bytearray(self.history)

    def __setstate__(self, ba):
        history = None
        core.DataStream.write_bytearray(ba, history)
        self.__init__(history)

    def get_items(self) -> List[webenginewidgets.WebEngineHistoryItem]:
        return [webenginewidgets.WebEngineHistoryItem(i) for i in self.history.items()]


if __name__ == "__main__":
    from prettyqt import widgets

    app = widgets.app()
    page = webenginewidgets.WebEnginePage()
    page.load_url("http://www.google.de")
    page.load_url("http://www.google.com")
    history = page.history()
    item = WebEngineHistory(history)
    print(len(item))
    app.main_loop()



# -*- coding: utf-8 -*-

import datetime
from typing import Optional

from qtpy import QtWidgets

from prettyqt import widgets
from prettyqt.utils import bidict, InvalidParamError, to_date


SELECTION_MODES = bidict(
    none=QtWidgets.QCalendarWidget.NoSelection,
    single=QtWidgets.QCalendarWidget.SingleSelection,
)

HEADER_FORMATS = bidict(
    none=QtWidgets.QCalendarWidget.NoVerticalHeader,
    week_numbers=QtWidgets.QCalendarWidget.ISOWeekNumbers,
)


QtWidgets.QCalendarWidget.__bases__ = (widgets.Widget,)


class CalendarWidget(QtWidgets.QCalendarWidget):
    def serialize_fields(self):
        return dict(date=self.get_date())

    def __setstate__(self, state):
        self.__init__()
        self.setSelectedDate(state["date"])

    def get_date(self) -> datetime.date:
        return to_date(self.selectedDate())

    def get_value(self) -> datetime.date:
        return self.get_date()

    def set_value(self, value):
        self.setSelectedDate(value)

    def set_selection_mode(self, mode: Optional[str]):
        """Set selection mode for given calendar widget.

        Allowed values are "single" or "none"

        Args:
            mode: selection mode to use

        Raises:
            InvalidParamError: mode does not exist
        """
        if mode is None:
            mode = "none"
        if mode not in SELECTION_MODES:
            raise InvalidParamError(mode, SELECTION_MODES)
        self.setSelectionMode(SELECTION_MODES[mode])

    def get_selection_mode(self) -> str:
        """Return current selection mode.

        Possible values: "single" or "none"

        Returns:
            selection mode
        """
        return SELECTION_MODES.inv[self.selectionMode()]


if __name__ == "__main__":
    app = widgets.app()
    w = CalendarWidget()
    w.show()
    app.main_loop()



# -*- coding: utf-8 -*-

from typing import Optional

from qtpy import QtWidgets

from prettyqt import core, widgets


class InputAndSlider(widgets.Widget):

    value_changed = core.Signal(int)

    def __init__(
        self, bounds: Optional[tuple] = None, parent: Optional[QtWidgets.QWidget] = None
    ):
        super().__init__(parent)
        self.path = None
        layout = widgets.BoxLayout("horizontal", self)
        layout.set_margin(0)
        self.spinbox = widgets.SpinBox()
        layout += self.spinbox
        self.slider = widgets.Slider()
        layout += self.slider
        if bounds:
            self.set_range(*bounds)
        self.spinbox.valueChanged.connect(self.slider.set_value)
        self.slider.valueChanged.connect(self.spinbox.set_value)
        self.spinbox.valueChanged.connect(self.value_changed)

    def serialize_fields(self):
        return dict(path=self.path)

    # def __setstate__(self, state):
    #     self.__init__(state["extensions"])
    #     self.set_path(state["path"])
    #     self.set_enabled(state.get("enabled", True))

    def set_range(self, min_val: int, max_val: int):
        self.spinbox.set_range(min_val, max_val)
        self.slider.set_range(min_val, max_val)

    def get_value(self) -> int:
        return self.spinbox.get_value()

    def set_value(self, value: int):
        self.spinbox.set_value(value)
        self.slider.set_value(value)

    def is_valid(self) -> bool:
        return self.spinbox.is_valid()

    def set_step_size(self, step_size: int):
        self.spinbox.set_step_size(step_size)
        self.slider.set_step_size(step_size)
        self.slider.setTickInterval(step_size)


if __name__ == "__main__":
    app = widgets.app()
    btn = InputAndSlider()
    btn.set_step_size(2)
    btn.slider.set_tick_position("below")
    btn.set_value(4)
    btn.show()
    btn.value_changed.connect(print)
    app.main_loop()



# -*- coding: utf-8 -*-

from typing import Optional

from qtpy import QtWidgets

from prettyqt import core, widgets


QtWidgets.QDial.__bases__ = (widgets.AbstractSlider,)


class Dial(QtWidgets.QDial):

    value_changed = core.Signal(int)

    def __init__(self, parent: Optional[QtWidgets.QWidget] = None):
        super().__init__(parent)
        self.valueChanged.connect(self.on_value_change)

    def serialize_fields(self):
        return dict(
            # notch_size=self.notchSize(),
            notch_target=self.notchTarget(),
            notches_visible=self.notchesVisible(),
            wrapping=self.wrapping(),
        )

    def __setstate__(self, state):
        self.__init__()
        self.set_range(*state["range"])
        self.set_value(state["value"])
        self.setToolTip(state.get("tool_tip", ""))
        self.setStatusTip(state.get("status_tip", ""))
        self.setEnabled(state.get("enabled", True))
        self.setSingleStep(state["single_step"])
        self.setPageStep(state["page_step"])
        self.setTracking(state["has_tracking"])
        self.setInvertedControls(state["inverted_controls"])
        self.setInvertedAppearance(state["inverted_appearance"])
        self.setNotchTarget(state["notch_target"])
        self.setNotchesVisible(state["notches_visible"])
        self.setWrapping(state["wrapping"])


if __name__ == "__main__":
    app = widgets.app()
    slider = Dial()
    slider.setRange(0, 100)
    slider.value_changed.connect(print)
    slider.show()
    app.main_loop()



# -*- coding: utf-8 -*-

from prettyqt import core, gui


class RegularExpressionValidator(gui.Validator):
    def __init__(self, regex=None):
        super().__init__()
        self.regex = None
        self.set_regex(regex)

    def __repr__(self):
        return f"RegularExpressionValidator(RegularExpression({self.get_regex()!r}))"

    def __getstate__(self):
        return dict(regexp=core.RegularExpression(self.regex))

    def __setstate__(self, state):
        self.__init__()
        self.regex = state["regexp"]

    def setRegularExpression(self, re):
        self.regex = re

    def regularExpression(self):
        return self.regex

    def set_regex(self, regex: str, flags: int = 0):
        self.regex = core.RegularExpression(regex, flags)

    def get_regex(self) -> str:
        if self.regex is None:
            raise TypeError()
        return self.regex.pattern()

    def validate(self, text: str, pos: int = 0) -> tuple:
        if self.regex is None:
            raise TypeError()
        if text == "":
            return (self.Intermediate, text, pos)
        match = self.regex.match(text, match_type="prefer_first")
        if match.hasPartialMatch():
            return (self.Intermediate, text, pos)
        if match.hasMatch():
            return (self.Acceptable, text, pos)
        return (self.Invalid, text, pos)


if __name__ == "__main__":
    from prettyqt import widgets

    app = widgets.app()
    w = widgets.LineEdit()
    val = RegularExpressionValidator()
    val.set_regex(r"\w\d\d")
    w.set_validator(val)
    w.show()
    app.main_loop()



# -*- coding: utf-8 -*-

from typing import Dict, List

from qtpy import QtNetwork

from prettyqt import network
from prettyqt.utils import bidict, helpers, InvalidParamError


KNOWN_HEADERS = network.networkrequest.KNOWN_HEADERS  # type: ignore


CAPABILITIES = bidict(
    tunneling=QtNetwork.QNetworkProxy.TunnelingCapability,
    listening=QtNetwork.QNetworkProxy.ListeningCapability,
    udp_tunneling=QtNetwork.QNetworkProxy.UdpTunnelingCapability,
    caching=QtNetwork.QNetworkProxy.CachingCapability,
    host_name_lookup=QtNetwork.QNetworkProxy.HostNameLookupCapability,
    sctp_tunneling=QtNetwork.QNetworkProxy.SctpTunnelingCapability,
    sctp_listening=QtNetwork.QNetworkProxy.SctpListeningCapability,
)

PROXY_TYPES = bidict(
    none=QtNetwork.QNetworkProxy.NoProxy,
    default=QtNetwork.QNetworkProxy.DefaultProxy,
    socks5=QtNetwork.QNetworkProxy.Socks5Proxy,
    http=QtNetwork.QNetworkProxy.HttpProxy,
    http_caching=QtNetwork.QNetworkProxy.HttpCachingProxy,
    ftp_caching=QtNetwork.QNetworkProxy.FtpCachingProxy,
)


class NetworkProxy(QtNetwork.QNetworkProxy):
    def get_capabilities(self) -> List[str]:
        return [k for k, v in CAPABILITIES.items() if v & self.capabilities()]

    def set_capabilities(self, *capability: str):
        for item in capability:
            if item not in CAPABILITIES:
                raise InvalidParamError(item, CAPABILITIES)
        flags = helpers.merge_flags(capability, CAPABILITIES)
        self.setCapabilities(flags)

    def get_header(self, name: str) -> str:
        if name not in KNOWN_HEADERS:
            raise InvalidParamError(name, KNOWN_HEADERS)
        return self.header(KNOWN_HEADERS[name])

    def set_header(self, name: str, value: str):
        if name not in KNOWN_HEADERS:
            raise InvalidParamError(name, KNOWN_HEADERS)
        self.setHeader(KNOWN_HEADERS[name], value)

    def get_headers(self) -> Dict[str, str]:
        return {
            bytes(h).decode(): bytes(self.rawHeader(h)).decode()
            for h in self.rawHeaderList()
        }

    def set_headers(self, headers: Dict[str, str]):
        for k, v in headers.items():
            self.setRawHeader(str.encode(k), str.encode(v))

    def set_type(self, typ: str):
        """Set proxy type.

        Valid values: "none", "default", "socks5", "http", "http_caching",
                      "ftp_caching"

        Args:
            typ: proxy type

        Raises:
            InvalidParamError: proxy type does not exist
        """
        if typ not in PROXY_TYPES:
            raise InvalidParamError(typ, PROXY_TYPES)
        self.setType(PROXY_TYPES[typ])

    def get_type(self) -> str:
        """Get the proxy type.

        Possible values: "none", "default", "socks5", "http", "http_caching",
                         "ftp_caching"

        Returns:
            type
        """
        return PROXY_TYPES.inv[self.type()]



# -*- coding: utf-8 -*-

from typing import Callable, Optional, Union
from dataclasses import dataclass

from prettyqt import constants
from prettyqt.utils import bidict

from qtpy import QtGui

SMALL_COL_WIDTH = 120
MEDIUM_COL_WIDTH = 200

ALIGNMENTS = bidict(
    left=constants.ALIGN_LEFT,
    right=constants.ALIGN_RIGHT,
    top=constants.ALIGN_TOP,
    bottom=constants.ALIGN_BOTTOM,
    top_left=constants.ALIGN_TOP_LEFT,
    top_right=constants.ALIGN_TOP_RIGHT,
    bottom_left=constants.ALIGN_BOTTOM_LEFT,
    bottom_right=constants.ALIGN_BOTTOM_RIGHT,
    center=constants.ALIGN_CENTER,
)


@dataclass(frozen=True)
class ColumnItem:
    """Determines how an object attribute is shown."""

    name: str
    label: Optional[Callable]
    checkstate: Optional[Callable] = None
    doc: str = "<no help available>"
    col_visible: bool = True
    width: int = SMALL_COL_WIDTH
    alignment: Optional[Union[Callable, int]] = None
    line_wrap: str = "none"
    foreground_color: Optional[Union[Callable, str]] = None
    background_color: Optional[Union[Callable, str]] = None
    decoration: Optional[Union[Callable, QtGui.QIcon]] = None
    font: Optional[Union[Callable, QtGui.QFont]] = None
    selectable: bool = True
    enabled: bool = True
    editable: bool = False
    checkable: bool = False
    tristate: bool = False

    def get_name(self):
        return self.name

    def get_flag(self):
        flag = constants.NO_FLAGS
        if self.selectable:
            flag |= constants.IS_SELECTABLE
        if self.enabled:
            flag |= constants.IS_ENABLED
        if self.editable:
            flag |= constants.IS_EDITABLE
        if self.checkable:
            flag |= constants.IS_CHECKABLE
        if self.tristate:
            flag |= constants.IS_USER_TRISTATE
        return flag

    def get_label(self, tree_item):
        if self.label is None:
            return ""
        elif callable(self.label):
            return self.label(tree_item)
        return self.label

    def get_checkstate(self, tree_item):
        if self.checkstate is None:
            return None
        elif callable(self.checkstate):
            return self.checkstate(tree_item)
        return self.checkstate

    def get_font(self, tree_item):
        if self.font is None:
            return None
        elif callable(self.font):
            return self.font(tree_item)
        return self.font

    def get_foreground_color(self, tree_item):
        if self.foreground_color is None:
            return None
        elif callable(self.foreground_color):
            return self.foreground_color(tree_item)
        return self.foreground_color

    def get_background_color(self, tree_item):
        if self.background_color is None:
            return None
        elif callable(self.background_color):
            return self.background_color(tree_item)
        return self.background_color

    def get_decoration(self, tree_item):
        if self.decoration is None:
            return None
        elif callable(self.decoration):
            return self.decoration(tree_item)
        return self.decoration

    def get_alignment(self, tree_item) -> int:
        if self.alignment is None:
            return constants.ALIGN_LEFT
        elif callable(self.alignment):
            return self.alignment(tree_item)
        elif isinstance(self.alignment, str):
            return ALIGNMENTS[self.alignment]
        return self.alignment



# -*- coding: utf-8 -*-

from qtpy import QtLocation

from prettyqt import location
from prettyqt.utils import bidict

VISIBILITY = bidict(
    unspecified=QtLocation.QLocation.UnspecifiedVisibility,
    device=QtLocation.QLocation.DeviceVisibility,
    private=QtLocation.QLocation.PrivateVisibility,
    public=QtLocation.QLocation.PublicVisibility,
)


class PlaceCategory(QtLocation.QPlaceCategory):
    def __str__(self):
        return self.name()

    def __bool__(self):
        return not self.isEmpty()

    def get_icon(self) -> location.PlaceIcon:
        return location.PlaceIcon(self.icon())

    def get_visibility(self) -> str:
        """Return the visibility of the place.

        Possible values are "unspecified", "device", "private", "public"

        Returns:
            Visibility
        """
        return VISIBILITY.inv[self.visibility()]


if __name__ == "__main__":
    segment = PlaceCategory()
    print(bool(segment))



# -*- coding: utf-8 -*-

# from qtpy import QtMultimedia
# from prettyqt import multimedia

# QtMultimedia.QAbstractPlanarVideoBuffer.__bases__ = (multimedia.AbstractVideoBuffer,)


# class AbstractPlanarVideoBuffer(QtMultimedia.QAbstractPlanarVideoBuffer):
#     pass



# -*- coding: utf-8 -*-

from typing import Optional

from qtpy import QtLocation

from prettyqt import positioning
from prettyqt.utils import bidict, InvalidParamError


INSTRUCTION_DIRECTION = bidict(
    none=QtLocation.QGeoManeuver.NoDirection,
    forward=QtLocation.QGeoManeuver.DirectionForward,
    bear_right=QtLocation.QGeoManeuver.DirectionBearRight,
    light_right=QtLocation.QGeoManeuver.DirectionLightRight,
    right=QtLocation.QGeoManeuver.DirectionRight,
    hard_right=QtLocation.QGeoManeuver.DirectionHardRight,
    u_turn_right=QtLocation.QGeoManeuver.DirectionUTurnRight,
    u_turn_left=QtLocation.QGeoManeuver.DirectionUTurnLeft,
    hard_left=QtLocation.QGeoManeuver.DirectionHardLeft,
    left=QtLocation.QGeoManeuver.DirectionLeft,
    light_left=QtLocation.QGeoManeuver.DirectionLightLeft,
    bear_left=QtLocation.QGeoManeuver.DirectionBearLeft,
)


class GeoManeuver(QtLocation.QGeoManeuver):
    def __bool__(self):
        return self.isValid()

    def __str__(self):
        return self.instructionText()

    def __setitem__(self, index: str, val):
        attrs = self.extendedAttributes()
        attrs[index] = val
        self.setExtendedAttributes(attrs)

    def __getitem__(self, index: str):
        attr = self.extendedAttributes()
        if index not in attr:
            raise KeyError(f"Key {index!r} does not exist.")
        return attr[index]

    def get_position(self) -> positioning.GeoCoordinate:
        return positioning.GeoCoordinate(self.position())

    def get_waypoint(self) -> Optional[positioning.GeoCoordinate]:
        wp = self.waypoint()
        if not wp.isValid():
            return None
        return positioning.GeoCoordinate(wp)

    def set_direction(self, direction: str):
        """Set the direction.

        Allowed values are "none", "forward", "bear_right", "light_right", "right",
                           "hard_right", "u_turn_right", "bear_left", "light_left",
                           "left", "hard_left", "u_turn_left"

        Args:
            direction: Direction

        Raises:
            InvalidParamError: direction does not exist
        """
        if direction not in INSTRUCTION_DIRECTION:
            raise InvalidParamError(direction, INSTRUCTION_DIRECTION)
        self.setDirection(INSTRUCTION_DIRECTION[direction])

    def get_direction(self) -> str:
        """Return current direction.

        Possible values: "none", "forward", "bear_right", "light_right", "right",
                         "hard_right", "u_turn_right", "bear_left", "light_left", "left",
                         "hard_left", "u_turn_left"

        Returns:
            Direction
        """
        return INSTRUCTION_DIRECTION.inv[self.direction()]


if __name__ == "__main__":
    maneuver = GeoManeuver()
    print(bool(maneuver))
    maneuver.setExtendedAttributes(dict(a="a"))
    maneuver["test"] = "test"
    print(bool(maneuver))
    waypoint = maneuver.get_waypoint()
    print(bool(waypoint))
    print(maneuver.extendedAttributes())



# -*- coding: utf-8 -*-

from typing import List, Optional

from qtpy import QtLocation

from prettyqt import location, positioning, core
from prettyqt.utils import bidict, InvalidParamError

TYPE = location.placecontent.TYPE  # type: ignore

VISIBILITY = bidict(
    unspecified=QtLocation.QLocation.UnspecifiedVisibility,
    device=QtLocation.QLocation.DeviceVisibility,
    private=QtLocation.QLocation.PrivateVisibility,
    public=QtLocation.QLocation.PublicVisibility,
)


class Place(QtLocation.QPlace):
    def __bool__(self):
        return not self.isEmpty()

    def __setitem__(self, index: str, val: Optional[QtLocation.QPlaceAttribute]):
        if val is None:
            val = location.PlaceAttribute()
        self.setExtendedAttribute(index, val)

    def __getitem__(self, index: str) -> location.PlaceAttribute:
        attr = self.extendedAttribute(index)
        return location.PlaceAttribute(attr)

    def get_categories(self) -> List[location.PlaceCategory]:
        return [location.PlaceCategory(i) for i in self.categories()]

    def get_contact_details(self, contact_type: str) -> List[location.PlaceContactDetail]:
        return [location.PlaceContactDetail(i) for i in self.contactDetails(contact_type)]

    def set_content(self, typ: str, value: str):
        if typ not in TYPE:
            raise InvalidParamError(typ, TYPE)
        self.setContent(TYPE[typ], value)

    def get_content(self, typ: str) -> str:
        if typ not in TYPE:
            raise InvalidParamError(typ, TYPE)
        return self.content(TYPE[typ])

    def get_icon(self) -> location.PlaceIcon:
        return location.PlaceIcon(self.icon())

    def get_location(self) -> positioning.GeoLocation:
        return positioning.GeoLocation(self.location())

    def get_primary_website(self) -> core.Url:
        return core.Url(self.primaryWebsite())

    def get_ratings(self) -> location.PlaceRatings:
        return location.PlaceRatings(self.ratings())

    def get_supplier(self) -> location.PlaceSupplier:
        return location.PlaceSupplier(self.supplier())

    def get_visibility(self) -> str:
        """Return visibility.

        Possible values are "unspecified", "device", "private", "public"

        Returns:
            Visibility
        """
        return VISIBILITY.inv[self.visibility()]


if __name__ == "__main__":
    place = Place()
    place["test"] = location.PlaceAttribute()
    assert place["test"] == "a"



# -*- coding: utf-8 -*-

from typing import List, Iterator, Union, Tuple

from qtpy import QtCore, QtGui

from prettyqt import core, gui
from prettyqt.utils import InvalidParamError

QtGui.QStandardItemModel.__bases__ = (core.AbstractItemModel,)

MATCH_FLAGS = dict(
    exact=QtCore.Qt.MatchExactly,
    contains=QtCore.Qt.MatchContains,
    starts_with=QtCore.Qt.MatchStartsWith,
    ends_with=QtCore.Qt.MatchEndsWith,
    wildcard=QtCore.Qt.MatchWildcard,
    regex=QtCore.Qt.MatchRegExp,
)


class StandardItemModel(QtGui.QStandardItemModel):
    def __getitem__(
        self, index: Union[int, Tuple[int, int], QtCore.QModelIndex]
    ) -> QtGui.QStandardItem:
        if isinstance(index, int):
            return self.item(index)
        elif isinstance(index, tuple):
            return self.item(*index)
        else:
            return self.itemFromIndex(index)

    def __delitem__(self, index: Union[int, Tuple[int, int]]):
        if isinstance(index, int):
            self.takeRow(index)
        elif isinstance(index, tuple):
            self.takeItem(*index)

    def __iter__(self) -> Iterator[QtGui.QStandardItem]:
        return iter(self.get_children())

    def __getstate__(self):
        return dict(items=self.get_children())

    def __setstate__(self, state):
        self.__init__()
        for item in state["items"]:
            self.appendRow(item)

    def __add__(self, other):
        if isinstance(other, (QtGui.QStandardItem, str)):
            self.add(other)
            return self
        raise TypeError("wrong type for addition")

    def get_children(self) -> List[QtGui.QStandardItem]:
        return [self.item(index) for index in range(self.rowCount())]

    def add(self, *item):
        for i in item:
            if isinstance(i, str):
                i = gui.StandardItem(i)
            self.appendRow(i)

    def find_items(
        self, text: str, column: int = 0, mode: str = "exact"
    ) -> List[QtGui.QStandardItem]:
        if mode not in MATCH_FLAGS:
            raise InvalidParamError(mode, MATCH_FLAGS)
        return self.findItems(text, MATCH_FLAGS[mode], column)


if __name__ == "__main__":
    import pickle

    from prettyqt import widgets

    model = gui.StandardItemModel()
    model.add("test")
    app = widgets.app()
    w = widgets.ListView()
    w.set_model(model)
    model += gui.StandardItem("Item")
    for item in model:
        pass
    with open("data.pkl", "wb") as jar:
        pickle.dump(model, jar)
    with open("data.pkl", "rb") as jar:
        model = pickle.load(jar)
    model += gui.StandardItem("Item2")
    w.show()
    app.main_loop()



# Based on: PySide examples/itemviews/simpletreemodel
# See: http://harmattan-dev.nokia.com/docs/library/html/qt4/itemviews-simpletreemodel.html

# Disabling the need for docstrings, all methods are tiny.
# pylint: disable=C0111

import logging

from prettyqt.utils import helpers

logger = logging.getLogger(__name__)

MAX_OBJ_STR_LEN = 50


def name_is_special(method_name):
    """Returns true if the method name starts and ends with two underscores."""
    return method_name.startswith("__") and method_name.endswith("__")


class TreeItem(object):
    """Tree node class that can be used to build trees of objects."""

    def __init__(self, obj, name, obj_path, is_attribute, parent=None):
        self.parent_item = parent
        self.obj = obj
        self.obj_name = str(name)
        self.obj_path = str(obj_path)
        self.is_attribute = is_attribute
        self.child_items = []
        self.has_children = True
        self.children_fetched = False

    def __repr__(self):
        if len(self.child_items) == 0:
            string = helpers.cut_off_str(self.obj, MAX_OBJ_STR_LEN)
            return f"<TreeItem(0x{id(self.obj):x}): {self.obj_path} = {string}>"
        else:
            n = len(self.child_items)
            return f"<TreeItem(0x{id(self.obj):x}): {self.obj_path} ({n:d} children)>"

    @property
    def is_special_attribute(self) -> bool:
        """Return true if the item represents a dunder attribute."""
        return self.is_attribute and name_is_special(self.obj_name)

    @property
    def is_callable_attribute(self) -> bool:
        """Return true if the items is an attribute and it is callable."""
        return self.is_attribute and callable(self.obj)

    def append_child(self, item):
        item.parent_item = self
        self.child_items.append(item)

    def insert_children(self, idx: int, items):
        self.child_items[idx:idx] = items
        for item in items:
            item.parent_item = self

    def child(self, row: int):
        return self.child_items[row]

    def child_count(self) -> int:
        return len(self.child_items)

    def parent(self):
        return self.parent_item

    def row(self) -> int:
        if self.parent_item:
            return self.parent_item.child_items.index(self)
        else:
            return 0

    def pretty_print(self, indent=0):
        logger.debug(indent * "    " + str(self))
        for child_item in self.child_items:
            child_item.pretty_print(indent + 1)



# -*- coding: utf-8 -*-

from typing import List

from qtpy import QtCore

from prettyqt import core


QtCore.QSortFilterProxyModel.__bases__ = (core.AbstractProxyModel,)


class SortFilterProxyModel(QtCore.QSortFilterProxyModel):

    HEADER: List[str] = []



# -*- coding: utf-8 -*-

from qtpy import QtWidgets

from prettyqt import widgets


QtWidgets.QStyleOptionRubberBand.__bases__ = (widgets.StyleOption,)


class StyleOptionRubberBand(QtWidgets.QStyleOptionRubberBand):
    pass



# -*- coding: utf-8 -*-

import contextlib
from typing import Callable

from qtpy import QtWidgets

from prettyqt import core, widgets


QtWidgets.QUndoStack.__bases__ = (core.Object,)


class UndoStack(QtWidgets.QUndoStack):
    def __len__(self) -> int:
        return self.count()

    def __getitem__(self, index: int) -> QtWidgets.QUndoCommand:
        return self.command(index)

    @contextlib.contextmanager
    def create_macro(self, text: str):
        self.beginMacro(text)
        yield None
        self.endMacro()

    def add_command(
        self, title: str, redo: Callable, undo: Callable
    ) -> widgets.UndoCommand:
        class MyCommand(widgets.UndoCommand):
            def redo(self):
                return redo()

            def undo(self):
                return undo()

        cmd = MyCommand(title)
        self.push(cmd)
        return cmd



from qtpy import QtWidgets, QtCore

from prettyqt import widgets, gui
from prettyqt.utils import bidict, InvalidParamError


FILL_RULES = bidict(odd_even=QtCore.Qt.OddEvenFill, winding=QtCore.Qt.WindingFill)

QtWidgets.QGraphicsPolygonItem.__bases__ = (widgets.AbstractGraphicsShapeItem,)


class GraphicsPolygonItem(QtWidgets.QGraphicsPolygonItem):
    def get_polygon(self):
        return gui.PolygonF(self.polygon())

    def serialize_fields(self):
        return dict(polygon=self.get_polygon(), fill_rule=self.get_fill_rule())

    def set_fill_rule(self, rule: str):
        if rule not in FILL_RULES:
            raise InvalidParamError(rule, FILL_RULES)
        self.setFillRule(FILL_RULES[rule])

    def get_fill_rule(self):
        return FILL_RULES.inv[self.fillRule()]



# -*- coding: utf-8 -*-

from typing import Iterator

import pathlib

from qtpy import QtCore

from prettyqt import core
from prettyqt.utils import bidict


COMPRESSION = bidict(
    none=QtCore.QResource.NoCompression,
    zlib=QtCore.QResource.ZlibCompression,
    zstd=QtCore.QResource.ZstdCompression,
)


class Resource(QtCore.QResource):
    def __repr__(self):
        return f"Resource({self.absoluteFilePath()!r}, {self.get_locale()!r})"

    def __bytes__(self):
        return bytes(self.uncompressedData())

    def __bool__(self):
        return self.isValid()

    def __iter__(self) -> Iterator[str]:
        return iter(self.children())

    def get_compression_algorithm(self) -> str:
        return COMPRESSION.inv[self.compressionAlgorithm()]

    def get_absolute_file_path(self) -> pathlib.Path:
        return pathlib.Path(self.absoluteFilePath())

    def get_locale(self) -> core.Locale:
        return core.Locale(self.locale())

    def get_last_modified(self) -> core.DateTime:
        return core.DateTime(self.lastModified())



# -*- coding: utf-8 -*-

from qtpy import QtCore

from prettyqt import core
from prettyqt.utils import bidict, InvalidParamError


CHILD_MODES = bidict(
    exclusive=QtCore.QState.ExclusiveStates, parallel=QtCore.QState.ParallelStates
)

RESTORE_POLICIES = bidict(
    dont_restore=QtCore.QState.DontRestoreProperties,
    restore=QtCore.QState.RestoreProperties,
)

QtCore.QState.__bases__ = (core.AbstractState,)


class State(QtCore.QState):
    def set_child_mode(self, mode: str):
        """Set child mode to use.

        Allowed values are "exclusive", "parallel"

        Args:
            mode: child mode to use

        Raises:
            InvalidParamError: child mode does not exist
        """
        if mode not in CHILD_MODES:
            raise InvalidParamError(mode, CHILD_MODES)
        self.setChildMode(CHILD_MODES[mode])

    def get_child_mode(self) -> str:
        """Return current child mode.

        Possible values: "exclusive", "parallel"

        Returns:
            child mode
        """
        return CHILD_MODES.inv[self.childMode()]


if __name__ == "__main__":
    reg = State()



# -*- coding: utf-8 -*-

from qtpy import QtCore

from prettyqt.utils import bidict

STRING_FORMATS = bidict(
    with_braces=QtCore.QUuid.WithBraces,
    without_braces=QtCore.QUuid.WithoutBraces,
    id_128=QtCore.QUuid.Id128,
)

VARIANTS = bidict(
    unknown=QtCore.QUuid.VarUnknown,
    ncs=QtCore.QUuid.NCS,
    dce=QtCore.QUuid.DCE,
    microsoft=QtCore.QUuid.Microsoft,
    reserved=QtCore.QUuid.Reserved,
)

VERSION = bidict(
    unknown=QtCore.QUuid.VerUnknown,
    time=QtCore.QUuid.Time,
    embedded_posix=QtCore.QUuid.EmbeddedPOSIX,
    name=QtCore.QUuid.Name,
    # md5=QtCore.QUuid.Md5,
    random=QtCore.QUuid.Random,
    sha1=QtCore.QUuid.Sha1,
)


class Uuid(QtCore.QUuid):
    def __repr__(self):
        return f"{self.__class__.__name__}({self.toString()!r})"

    def __bool__(self):
        return not self.isNull()

    def get_variant(self) -> str:
        return VARIANTS.inv[self.variant()]

    def get_version(self) -> str:
        return VERSION.inv[self.version()]

    @classmethod
    def create_uuid(cls):
        return cls(cls.createUuid())


if __name__ == "__main__":
    address = Uuid.create_uuid()
    print(repr(address))



# -*- coding: utf-8 -*-

from typing import Union

from qtpy import QtGui

from prettyqt import gui
from prettyqt.utils import bidict, colors, InvalidParamError


FONT_PROPERTY_INHERITANCE_BEHAVIOUR = bidict(
    none=QtGui.QTextCharFormat.FontPropertiesSpecifiedOnly,
    single=QtGui.QTextCharFormat.FontPropertiesAll,
)

UNDERLINE_STYLES = bidict(
    none=QtGui.QTextCharFormat.NoUnderline,
    single=QtGui.QTextCharFormat.SingleUnderline,
    dash=QtGui.QTextCharFormat.DashUnderline,
    dot=QtGui.QTextCharFormat.DotLine,
    dashdot=QtGui.QTextCharFormat.DashDotLine,
    dashdotline=QtGui.QTextCharFormat.DashDotDotLine,
    wave=QtGui.QTextCharFormat.WaveUnderline,
    spellcheck=QtGui.QTextCharFormat.SpellCheckUnderline,
)

VERTICAL_ALIGNMENT = bidict(
    normal=QtGui.QTextCharFormat.AlignNormal,
    super_script=QtGui.QTextCharFormat.AlignSuperScript,
    sub_script=QtGui.QTextCharFormat.AlignSubScript,
    middle=QtGui.QTextCharFormat.AlignMiddle,
    bottom=QtGui.QTextCharFormat.AlignBottom,
    top=QtGui.QTextCharFormat.AlignTop,
    baseline=QtGui.QTextCharFormat.AlignBaseline,
)


QtGui.QTextCharFormat.__bases__ = (gui.TextFormat,)


class TextCharFormat(QtGui.QTextCharFormat):
    def __init__(
        self,
        text_color: Union[colors.ColorType, QtGui.QBrush] = None,
        bold: bool = False,
        italic: bool = False,
    ):
        super().__init__()
        if text_color is not None:
            self.set_foreground_color(text_color)
        if bold:
            self.set_font_weight("bold")
        self.setFontItalic(italic)

    def set_foreground_color(self, color: Union[colors.ColorType, QtGui.QBrush]):
        if not isinstance(color, QtGui.QBrush):
            color = colors.get_color(color)
        self.setForeground(color)

    def set_background_color(self, color: Union[colors.ColorType, QtGui.QBrush]):
        if not isinstance(color, QtGui.QBrush):
            color = colors.get_color(color)
        self.setBackground(color)

    def set_font_weight(self, weight: str):
        """Set the font weight.

        Valid values are "thin", "extra_light", light", "medium", "demi_bold", "bold",
                         "extra_bold", normal", "black"

        Args:
            weight: font weight

        Raises:
            InvalidParamError: invalid font weight
        """
        if weight not in gui.font.WEIGHTS:
            raise InvalidParamError(weight, gui.font.WEIGHTS)
        self.setFontWeight(gui.font.WEIGHTS[weight])

    def get_font_weight(self) -> str:
        """Get current font weight.

        Possible values are "thin", "light", "medium" or "bold"

        Returns:
            current font weight
        """
        return gui.font.WEIGHTS.inv[self.fontWeight()]

    def set_underline_style(self, style: str):
        """Set the underline style.

        Valid values are "none", "single", "dash", "dot", "dashdot", "dashdotline",
        "wave", "spellcheck"

        Args:
            style: underline style

        Raises:
            InvalidParamError: invalid underline style
        """
        if style not in UNDERLINE_STYLES:
            raise InvalidParamError(style, UNDERLINE_STYLES)
        self.setUnderlineStyle(UNDERLINE_STYLES[style])

    def get_underline_style(self) -> str:
        """Get current underline style.

        Possible values are "none", "single", "dash", "dot", "dashdot", "dashdotline",
        "wave", "spellcheck"

        Returns:
            current underline style
        """
        return UNDERLINE_STYLES.inv[self.underlineStyle()]

    def set_vertical_alignment(self, alignment: str):
        """Set the vertical alignment.

        Valid values: "normal", "super_script", "sub_script", "middle", "bottom", "top",
        "baseline"

        Args:
            alignment: vertical alignment

        Raises:
            InvalidParamError: invalid vertical alignment
        """
        if alignment not in VERTICAL_ALIGNMENT:
            raise InvalidParamError(alignment, VERTICAL_ALIGNMENT)
        self.setVerticalAlignment(VERTICAL_ALIGNMENT[alignment])

    def get_vertical_alignment(self) -> str:
        """Get current vertical alignment.

        Possible values: "normal", "super_script", "sub_script", "middle", "bottom",
        "top", "baseline"

        Returns:
            current vertical alignment
        """
        return VERTICAL_ALIGNMENT.inv[self.verticalAlignment()]

    def set_font_style_hint(self, hint: str):
        """Set the font style hint.

        Valid values are "any", "sans_serif", "serif", "typewriter", "decorative",
        "monospace", "fantasy", "cursive", "system"

        Args:
            hint: font style hint

        Raises:
            InvalidParamError: invalid font style hint
        """
        if hint not in gui.font.STYLE_HINTS:
            raise InvalidParamError(hint, gui.font.STYLE_HINTS)
        self.setFontStyleHint(gui.font.STYLE_HINTS[hint])

    def select_full_width(self, value: bool = True):
        self.setProperty(QtGui.QTextFormat.FullWidthSelection, value)

    def get_font(self) -> gui.Font:
        return gui.Font(self.font())



# -*- coding: utf-8 -*-

from typing import List

from qtpy import QtWidgets

from prettyqt import core, widgets
from prettyqt.utils import bidict


CHANGE_FLAGS = bidict(
    scale_factor=QtWidgets.QPinchGesture.ScaleFactorChanged,
    rotation_angle=QtWidgets.QPinchGesture.RotationAngleChanged,
    center_point=QtWidgets.QPinchGesture.CenterPointChanged,
)

QtWidgets.QPinchGesture.__bases__ = (widgets.Gesture,)


class PinchGesture(QtWidgets.QPinchGesture):
    def get_start_center_point(self) -> core.PointF:
        return core.PointF(self.startCenterPoint())

    def get_center_point(self) -> core.PointF:
        return core.PointF(self.centerPoint())

    def get_last_center_point(self) -> core.PointF:
        return core.PointF(self.lastCenterPoint())

    def get_change_flags(self) -> List[str]:
        return [k for k, v in CHANGE_FLAGS.items() if v & self.changeFlags()]

    def set_change_flags(self, **kwargs):
        flag = QtWidgets.QPinchGesture.ChangeFlags(0)
        for k, v in kwargs.items():
            if v is True:
                flag |= CHANGE_FLAGS[k]
        self.setChangeFlags(flag)

    def get_total_change_flags(self) -> List[str]:
        return [k for k, v in CHANGE_FLAGS.items() if v & self.totalChangeFlags()]

    def set_total_change_flags(self, **kwargs):
        flag = QtWidgets.QPinchGesture.ChangeFlags(0)
        for k, v in kwargs.items():
            if v is True:
                flag |= CHANGE_FLAGS[k]
        self.setTotalChangeFlags(flag)


if __name__ == "__main__":
    from prettyqt import widgets

    app = widgets.app()
    gesture = PinchGesture()



# -*- coding: utf-8 -*-

from typing import Optional, Union
import pathlib

from qtpy import QtWidgets

from prettyqt import widgets


class ImageViewer(widgets.Widget):
    def __init__(
        self,
        title: str = "",
        path: Union[pathlib.Path, str] = None,
        parent: Optional[QtWidgets.QWidget] = None,
    ):
        super().__init__(parent)
        if title:
            self.set_title(title)
        self.image = None
        if path:
            self.image = widgets.Label.image_from_path(path, parent=self)
        self.show()


if __name__ == "__main__":
    app = widgets.app()
    ex = ImageViewer()
    app.main_loop()



# -*- coding: utf-8 -*-

from qtpy import PYQT5, PYSIDE2

if PYQT5:
    from PyQt5 import QtWebEngineCore  # type: ignore
elif PYSIDE2:
    from PySide2 import QtWebEngineCore

from prettyqt import core


QtWebEngineCore.QWebEngineUrlSchemeHandler.__bases__ = (core.Object,)


class WebEngineUrlSchemeHandler(QtWebEngineCore.QWebEngineUrlSchemeHandler):
    pass


if __name__ == "__main__":
    item = WebEngineUrlSchemeHandler()



# -*- coding: utf-8 -*-

import contextlib
import pathlib
from typing import List, Mapping, Optional, Union, Dict, Any, Iterator
import logging

from qtpy import QtCore

from prettyqt import core
from prettyqt.utils import bidict, InvalidParamError

logger = logging.getLogger(__name__)

FORMATS = bidict(native=QtCore.QSettings.NativeFormat, ini=QtCore.QSettings.IniFormat)

SCOPES = bidict(user=QtCore.QSettings.UserScope, system=QtCore.QSettings.SystemScope)


QtCore.QSettings.__bases__ = (core.Object,)


class Settings(QtCore.QSettings):
    def __init__(self, *args, settings_id: Optional[str] = None):
        super().__init__(*args)
        self.settings_id = settings_id

    def __repr__(self):
        return f"Settings: {self.as_dict()}"

    def __contains__(self, key: str) -> bool:
        return self.contains(key)

    def __enter__(self):
        if self.settings_id:
            self.beginGroup(self.settings_id)
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        if self.settings_id:
            self.endGroup()

    def __getitem__(self, index: str):
        return self.get_value(index)

    def __setitem__(self, name: str, value):
        return self.set_value(name, value)

    def __delitem__(self, key: str):
        if not self.contains(key):
            raise KeyError(key)
        return self.remove(key)

    def __iter__(self) -> Iterator[str]:
        return iter(self.allKeys())

    def __len__(self) -> int:
        return len(self.allKeys())

    @classmethod
    def build_from_dict(cls, dct: Dict[str, Any]):
        settings = cls()
        for k, v in dct.items():
            settings.set_value(k, v)
        return settings

    def as_dict(self) -> Dict[str, Any]:
        return {k: v for k, v in self.items()}

    def set_value(self, key: str, value):
        if not self.applicationName():
            raise RuntimeError("no app name defined")
        self.setValue(key, dict(value=value))

    def set_values(self, dct: Dict[str, Any]):
        for k, v in dct.items():
            self.set_value(k, v)

    def get_value(self, key: str, default=None):
        if not self.contains(key):
            return default
        val = self.value(key)
        # this is for migration
        if not isinstance(val, dict) or "value" not in val:
            self.set_value(key, val)
            return val
        return val["value"]

    @classmethod
    def set_default_format(cls, fmt: str):
        """Set the default format.

        possible values are "native", "ini"

        Args:
            fmt: the default format to use

        Raises:
            InvalidParamError: invalid format
        """
        if fmt not in FORMATS:
            raise InvalidParamError(fmt, FORMATS)
        cls.setDefaultFormat(FORMATS[fmt])

    @classmethod
    def get_default_format(cls) -> str:
        """Return default settings format.

        possible values are "native", "ini"

        Returns:
            default settings format
        """
        return FORMATS.inv[cls.defaultFormat()]

    def get_scope(self) -> str:
        """Return scope.

        possible values are "user", "system"

        Returns:
            scope
        """
        return SCOPES.inv[self.scope()]

    @classmethod
    def set_path(cls, fmt: str, scope: str, path: Union[str, pathlib.Path]):
        """Set the path to the settings file.

        Args:
            fmt: the default format to use
            scope: the scope to use

        Raises:
            InvalidParamError: invalid format or scope
        """
        if fmt not in FORMATS:
            raise InvalidParamError(fmt, FORMATS)
        if scope not in SCOPES:
            raise InvalidParamError(scope, SCOPES)
        cls.setPath(FORMATS[fmt], SCOPES[scope], str(path))

    @contextlib.contextmanager
    def group(self, prefix: str):
        """Context manager for setting groups.

        Args:
            prefix: setting prefix for group
        """
        self.beginGroup(prefix)
        yield None
        self.endGroup()

    @contextlib.contextmanager
    def write_array(self, prefix: str, size: int = -1):
        """Context manager for writing arrays.

        Args:
            prefix: prefix for settings array
            size: size of settings array
        """
        self.beginWriteArray(prefix, size)
        yield None
        self.endArray()

    @contextlib.contextmanager
    def read_array(self, prefix: str):
        """Context manager for reading arrays.

        Args:
            prefix: prefix for settings array
        """
        self.beginReadArray(prefix)
        yield None
        self.endArray()

    # Dictionary interface

    def get(self, key: str, default=None):
        return self.get_value(key, default)

    def setdefault(self, key: str, default=None):
        if not self.contains(key):
            self.set_value(key, default)
            return default
        return self.get_value(key)

    def keys(self) -> List[str]:
        return self.allKeys()

    def values(self):
        return (self.get_value(key) for key in self.allKeys())

    def items(self):
        return zip(self.keys(), self.values())

    def pop(self, key: str):
        if self.contains(key):
            return self.get_value(key)
        raise KeyError(key)

    def popitem(self) -> tuple:
        key = self.keys()[0]
        return (key, self.get_value(key))

    def update(self, other: Mapping):
        for k, v in other.items():
            self.set_value(k, v)


def register_extensions(
    *exts: List[str],
    app_name: Optional[str] = None,
    app_path: Union[None, str, pathlib.Path] = None,
):
    logger.debug(f"assigning extensions {exts} to {app_name}")
    s = Settings("HKEY_CURRENT_USER\\SOFTWARE\\Classes", Settings.NativeFormat)
    if app_path is None:
        app_path = core.Dir.toNativeSeparators(core.CoreApplication.applicationFilePath())
    if app_name is None:
        app_name = core.CoreApplication.applicationName()
    for ext in exts:
        s.setValue(f"{ext}/DefaultIcon/.", app_path)  # perhaps ,0 after app_path
        s.setValue(f"{ext}/.", app_name)
    s.setValue(f"{app_name}/shell/open/command/.", f"{app_path} %1")


if __name__ == "__main__":
    settings = Settings("1", "2")
    settings["1"] = True
    print(settings["1"])
    print(type(settings.get("1")))
    del settings["hallo"]



# -*- coding: utf-8 -*-

import contextlib
from typing import Union

from qtpy import QtGui

from prettyqt.utils import bidict


MOVE_MODES = bidict(move=QtGui.QTextCursor.MoveAnchor, keep=QtGui.QTextCursor.KeepAnchor)

MOVE_OPERATIONS = bidict(
    no_move=QtGui.QTextCursor.NoMove,
    start=QtGui.QTextCursor.Start,
    start_of_line=QtGui.QTextCursor.StartOfLine,
    start_of_block=QtGui.QTextCursor.StartOfBlock,
    start_of_word=QtGui.QTextCursor.StartOfWord,
    previous_block=QtGui.QTextCursor.PreviousBlock,
    previous_char=QtGui.QTextCursor.PreviousCharacter,
    previous_word=QtGui.QTextCursor.PreviousWord,
    up=QtGui.QTextCursor.Up,
    left=QtGui.QTextCursor.Left,
    word_left=QtGui.QTextCursor.WordLeft,
    end=QtGui.QTextCursor.End,
    end_of_line=QtGui.QTextCursor.EndOfLine,
    end_of_word=QtGui.QTextCursor.EndOfWord,
    end_of_block=QtGui.QTextCursor.EndOfBlock,
    next_block=QtGui.QTextCursor.NextBlock,
    next_char=QtGui.QTextCursor.NextCharacter,
    next_word=QtGui.QTextCursor.NextWord,
    down=QtGui.QTextCursor.Down,
    right=QtGui.QTextCursor.Right,
    word_right=QtGui.QTextCursor.WordRight,
    next_cell=QtGui.QTextCursor.NextCell,
    previous_cell=QtGui.QTextCursor.PreviousCell,
    next_row=QtGui.QTextCursor.NextRow,
    previous_row=QtGui.QTextCursor.PreviousRow,
)

SELECTION_TYPES = bidict(
    document=QtGui.QTextCursor.Document,
    block_under_cursor=QtGui.QTextCursor.BlockUnderCursor,
    line_under_cursor=QtGui.QTextCursor.LineUnderCursor,
    word_under_cursor=QtGui.QTextCursor.WordUnderCursor,
)


class TextCursor(QtGui.QTextCursor):
    def move_position(self, operation: str, mode: str = "move", n: int = 1) -> bool:
        op = MOVE_OPERATIONS[operation]
        mode = MOVE_MODES[mode]
        return self.movePosition(op, mode, n)

    def set_position(self, pos: int, mode: str = "move"):
        """Set cursor to given position.

        Args:
            pos: Cursor position
            mode: Move mode
        """
        self.setPosition(pos, MOVE_MODES[mode])

    def select(self, selection):
        if selection in SELECTION_TYPES:
            selection = SELECTION_TYPES[selection]
        super().select(selection)

    def span(self) -> tuple:
        return (self.anchor(), self.position())

    def select_text(self, start_pos: Union[int, str], end_pos: Union[int, str]) -> str:
        """Select text from start position to end position.

        Positions can be either an integer index or a move operation

        Args:
            start_pos: Start position
            end_pos: End position
        """
        if isinstance(start_pos, int):
            self.set_position(start_pos)
        else:
            self.move_position(start_pos)
        if isinstance(end_pos, int):
            self.set_position(end_pos, mode="keep")
        else:
            self.move_position(end_pos, mode="keep")
        return self.selectedText()

    def replace_text(self, start_pos: int, end_pos: Union[str, int], to_replace: str):
        if isinstance(start_pos, int):
            self.set_position(start_pos)
        else:
            self.move_position(start_pos)
        if isinstance(end_pos, int):
            self.set_position(end_pos, mode="keep")
        else:
            self.move_position(end_pos, mode="keep")
        self.insertText(to_replace)
        self.select_text(start_pos, start_pos + len(to_replace))

    @contextlib.contextmanager
    def edit_block(self):
        """Context manager for edit blocks. Can be used for undo actions."""
        self.beginEditBlock()
        yield
        self.endEditBlock()



# -*- coding: utf-8 -*-

import pathlib

from qtpy import QtCore

from prettyqt import core
from prettyqt.utils import bidict, InvalidParamError

LOCATIONS = bidict(
    prefix=QtCore.QLibraryInfo.PrefixPath,
    documentation=QtCore.QLibraryInfo.DocumentationPath,
    headers=QtCore.QLibraryInfo.HeadersPath,
    libraries=QtCore.QLibraryInfo.LibrariesPath,
    library_executables=QtCore.QLibraryInfo.LibraryExecutablesPath,
    binaries=QtCore.QLibraryInfo.BinariesPath,
    plugins=QtCore.QLibraryInfo.PluginsPath,
    imports=QtCore.QLibraryInfo.ImportsPath,
    qml2_imports=QtCore.QLibraryInfo.Qml2ImportsPath,
    arch_data=QtCore.QLibraryInfo.ArchDataPath,
    data=QtCore.QLibraryInfo.DataPath,
    translations=QtCore.QLibraryInfo.TranslationsPath,
    examples=QtCore.QLibraryInfo.ExamplesPath,
    tests=QtCore.QLibraryInfo.TestsPath,
    settings=QtCore.QLibraryInfo.SettingsPath,
)


class LibraryInfo(QtCore.QLibraryInfo):
    def __class_getitem__(cls, name: str) -> pathlib.Path:
        return cls.get_location(name)

    @classmethod
    def get_location(cls, location: str) -> pathlib.Path:
        if location not in LOCATIONS:
            raise InvalidParamError(location, LOCATIONS)
        return pathlib.Path(cls.location(LOCATIONS[location]))

    @classmethod
    def get_version(cls) -> core.VersionNumber:
        return core.VersionNumber(cls.version())


if __name__ == "__main__":
    path = LibraryInfo.get_location("prefix")
    print(path)



# -*- coding: utf-8 -*-

from typing import Union, Optional

from qtpy import QtCore, QtWidgets

from prettyqt import core, widgets
from prettyqt.utils import bidict, InvalidParamError


MODES = bidict(
    default=QtWidgets.QLayout.SetDefaultConstraint,
    fixed=QtWidgets.QLayout.SetFixedSize,
    minimum=QtWidgets.QLayout.SetMinimumSize,
    maximum=QtWidgets.QLayout.SetMaximumSize,
    min_and_max=QtWidgets.QLayout.SetMinAndMaxSize,
    none=QtWidgets.QLayout.SetNoConstraint,
)

ALIGNMENTS = bidict(
    left=QtCore.Qt.AlignLeft,
    right=QtCore.Qt.AlignRight,
    top=QtCore.Qt.AlignTop,
    bottom=QtCore.Qt.AlignBottom,
)


QtWidgets.QLayout.__bases__ = (core.Object, widgets.LayoutItem)


class Layout(QtWidgets.QLayout):
    def __getitem__(
        self, index: Union[str, int]
    ) -> Optional[Union[QtWidgets.QWidget, QtWidgets.QLayout]]:
        if isinstance(index, int):
            item = self.itemAt(index)
            widget = item.widget()
            if widget is None:
                widget = item.layout()
        elif isinstance(index, str):
            widget = self.find_child(typ=QtCore.QObject, name=index)
        return widget

    def __delitem__(self, index: int):
        item = self.itemAt(index)
        self.removeItem(item)

    def __len__(self) -> int:
        return self.count()

    def __repr__(self):
        return f"{self.__class__.__name__}()"

    def __iter__(self):
        return iter(self[i] for i in range(self.count()))

    def __contains__(self, item: Union[QtWidgets.QWidget, QtWidgets.QLayout]):
        return item in self.get_children()

    def serialize_fields(self):
        return dict(
            size_mode=self.get_size_mode(),
            spacing=self.spacing(),
            enabled=self.isEnabled(),
        )

    def get_children(self) -> list:
        return list(self)

    def set_margin(self, margin: int):
        self.setContentsMargins(margin, margin, margin, margin)

    def set_spacing(self, pixels: int):
        self.setSpacing(pixels)

    def set_size_mode(self, mode: str):
        """Set the size mode of the layout.

        Allowed values are "default", "fixed", "minimum", "maximum", "min_and_max", "none"

        Args:
            mode: size mode for the layout

        Raises:
            InvalidParamError: size mode does not exist
        """
        if mode not in MODES:
            raise InvalidParamError(mode, MODES)
        self.setSizeConstraint(MODES[mode])

    def get_size_mode(self) -> str:
        """Return current size mode.

        Possible values: "default", "fixed", "minimum", "maximum", "min_and_max", "none"

        Returns:
            size mode
        """
        return MODES.inv[self.sizeConstraint()]

    def set_alignment(
        self,
        alignment: str,
        item: Optional[Union[QtWidgets.QWidget, QtWidgets.QLayout]] = None,
    ):
        """Set the alignment for widget / layout to alignment.

        Returns true if w is found in this layout (not including child layouts).

        Allowed values for alignment:  "left", "right", "top", "bottom"

        Args:
            alignment: alignment for the layout
            item: set alignment for specific child only

        Raises:
            InvalidParamError: alignment does not exist
        """
        if alignment not in ALIGNMENTS:
            raise InvalidParamError(alignment, ALIGNMENTS)
        if item is not None:
            return self.setAlignment(item, ALIGNMENTS[alignment])
        else:
            return self.setAlignment(ALIGNMENTS[alignment])

    def add(self, *item: Union[QtWidgets.QWidget, QtWidgets.QLayout]):
        for i in item:
            if isinstance(i, QtWidgets.QWidget):
                self.addWidget(i)
            elif isinstance(i, QtWidgets.QLayout):
                w = widgets.Widget()
                w.set_layout(i)
                self.addWidget(w)
            else:
                raise TypeError("add_item only supports widgets and layouts")



# -*- coding: utf-8 -*-

import functools
from typing import Iterable, Optional, Union, List

from deprecated import deprecated
from qtpy import QtCore, QtWidgets

from prettyqt import core, widgets
from prettyqt.utils import bidict, InvalidParamError


ORIENTATIONS = bidict(horizontal=QtCore.Qt.Horizontal, vertical=QtCore.Qt.Vertical)

MODES = bidict(
    interactive=QtWidgets.QHeaderView.Interactive,
    fixed=QtWidgets.QHeaderView.Fixed,
    stretch=QtWidgets.QHeaderView.Stretch,
    resize_to_contents=QtWidgets.QHeaderView.ResizeToContents,
)


QtWidgets.QHeaderView.__bases__ = (widgets.AbstractItemView,)


class HeaderView(QtWidgets.QHeaderView):

    section_vis_changed = QtCore.Signal(int, bool)

    def __init__(
        self, orientation: Union[str, int], parent: Optional[QtWidgets.QWidget] = None
    ):
        if orientation in ORIENTATIONS:
            orientation = ORIENTATIONS[orientation]
        super().__init__(orientation, parent=parent)
        self.setSectionsMovable(True)
        self.setSectionsClickable(True)
        self._widget_name = parent.id if parent is not None else ""

    def save_state(
        self, settings: Optional[core.Settings] = None, key: Optional[str] = None
    ):
        settings = core.Settings() if settings is None else settings
        key = f"{self._widget_name}.state" if key is None else key
        settings.set_value(key, self.saveState())

    def load_state(
        self, settings: Optional[core.Settings] = None, key: Optional[str] = None
    ) -> bool:
        settings = core.Settings() if settings is None else settings
        key = f"{self._widget_name}.state" if key is None else key
        state = settings.get(key, None)
        if state is not None:
            self.restoreState(state)
            return True
        return False

    def resize_sections(self, mode: str):
        self.resizeSections(MODES[mode])

    @deprecated(reason="This method is deprecated, use set_resize_mode instead.")
    def resize_mode(self, mode: str, col: Optional[int] = None):
        self.set_resize_mode(mode, col)

    def set_resize_mode(self, mode: str, col: Optional[int] = None):
        if mode not in MODES:
            raise InvalidParamError(mode, MODES)
        if col is None:
            self.setSectionResizeMode(MODES[mode])
        else:
            self.setSectionResizeMode(col, MODES[mode])

    def get_section_labels(self) -> List[str]:
        model = self.model()
        return [
            model.headerData(i, QtCore.Qt.Horizontal, QtCore.Qt.DisplayRole)
            for i in range(self.count())
        ]

    def contextMenuEvent(self, event):
        """Context menu for our files tree."""
        menu = widgets.Menu(parent=self)
        actions = self.get_header_actions()
        menu.add_actions(actions)
        menu.exec_(self.mapToGlobal(event.pos()))

    def get_header_actions(self) -> List[widgets.Action]:
        actions = list()
        labels = self.get_section_labels()[1:]
        for i, header_label in enumerate(labels, start=1):
            val = not self.isSectionHidden(i)
            action = widgets.Action(text=header_label, checkable=True, checked=val)
            fn = functools.partial(self.set_section_hidden, i=i, hide=val)
            action.triggered.connect(fn)
            actions.append(action)
        return actions

    def set_section_hidden(self, i: int, hide: bool):
        self.section_vis_changed.emit(i, hide)
        self.setSectionHidden(i, hide)

    def set_sizes(self, sizes: Iterable[Optional[int]]):
        for i, size in enumerate(sizes):
            if size is not None:
                self.resizeSection(i, size)

    def set_default_section_size(self, size: Optional[int]):
        if size is None:
            self.resetDefaultSectionSize()
        else:
            self.setDefaultSectionSize(size)

    def stretch_last_section(self, stretch: bool = True):
        self.setStretchLastSection(stretch)


if __name__ == "__main__":
    app = widgets.app()
    header = HeaderView("horizontal")
    header.show()
    app.main_loop()



# -*- coding: utf-8 -*-
from typing import List

from qtpy import PYQT5, PYSIDE2

if PYQT5:
    from PyQt5 import QtPositioning
elif PYSIDE2:
    from PySide2 import QtPositioning


from prettyqt import positioning

QtPositioning.QGeoPath.__bases__ = (positioning.GeoShape,)


class GeoPath(QtPositioning.QGeoPath):
    def __len__(self):
        return self.size()

    def __getitem__(self, index: int) -> positioning.GeoCoordinate:
        return positioning.GeoCoordinate(self.coordinateAt(index))

    def __setitem__(self, index: int, value: QtPositioning.QGeoCoordinate):
        self.replaceCoordinate(index, value)

    def __delitem__(self, index: int):
        self.removeCoordinate(index)

    def __add__(self, other: QtPositioning.QGeoCoordinate):
        self.addCoordinate(other)
        return self

    def __repr__(self):
        p = ", ".join([f"{p!r}" for p in self.get_path()])
        return f"GeoPath({p})"

    def get_path(self) -> List[positioning.GeoCoordinate]:
        return [positioning.GeoCoordinate(p) for p in self.path()]


if __name__ == "__main__":
    path = GeoPath()
    print(str(path))



# -*- coding: utf-8 -*-

from __future__ import annotations

import pathlib
from typing import Union, Optional, Dict, Tuple, List


import qtawesome as qta
from qtpy import QtGui, QtCore

from prettyqt import core, gui
from prettyqt.utils import bidict, InvalidParamError

MODES = bidict(
    normal=QtGui.QIcon.Normal,
    disabled=QtGui.QIcon.Disabled,
    active=QtGui.QIcon.Active,
    selected=QtGui.QIcon.Selected,
)

STATES = bidict(off=QtGui.QIcon.Off, on=QtGui.QIcon.On)

IconType = Union[QtGui.QIcon, str, pathlib.Path, None]

key_type = Tuple[Optional[str], Optional[str], bool]
icon_cache: Dict[key_type, QtGui.QIcon] = dict()


def get_icon(
    icon: IconType, color: Optional[str] = None, as_qicon: bool = False
) -> QtGui.QIcon:
    """Get icon with given color.

    Qtawesome already caches icons, but since we construct our own subclassed icon,
    we cache, too.
    """
    if isinstance(icon, QtGui.QIcon):
        return icon if as_qicon else Icon(icon)
    if isinstance(icon, pathlib.Path):
        icon = str(icon)
    if (icon, color, as_qicon) in icon_cache:
        return icon_cache[(icon, color, as_qicon)]
    if isinstance(icon, str) and icon.startswith("mdi."):
        if color is not None:
            new = qta.icon(icon, color=color)
        else:
            new = qta.icon(icon)
    else:
        new = QtGui.QIcon(icon)
    icon = new if as_qicon else Icon(new)
    icon_cache[(icon, color, as_qicon)] = icon
    return icon


def set_defaults(*args, **kwargs):
    qta.set_defaults(*args, **kwargs)


class Icon(QtGui.QIcon):
    def __repr__(self):
        return f"{self.__class__.__name__}()"

    def __bool__(self):
        return not bool(self.isNull())

    def __getstate__(self):
        pixmap = self.pixmap(256, 256)
        return core.DataStream.create_bytearray(pixmap)

    def __setstate__(self, ba):
        px = QtGui.QPixmap()
        core.DataStream.write_bytearray(ba, px)
        super().__init__(px)

    @classmethod
    def for_color(cls, color_str: str) -> Icon:
        color = gui.Color.from_text(color_str)
        if color.isValid():
            bitmap = gui.Pixmap(16, 16)
            bitmap.fill(color)
            icon = cls(bitmap)
        else:
            icon = cls(qta.icon("mdi.card-outline"))
        return icon

    @classmethod
    def from_image(cls, image: QtGui.QImage):
        return cls(gui.Pixmap.fromImage(image))

    def get_available_sizes(
        self, mode: str = "normal", state: str = "off"
    ) -> List[core.Size]:
        if mode not in MODES:
            raise InvalidParamError(mode, MODES)
        if state not in STATES:
            raise InvalidParamError(state, STATES)
        return [core.Size(i) for i in self.availableSizes(MODES[mode], STATES[state])]

    def add_pixmap(self, data: Union[QtCore.QByteArray, QtGui.QPixmap, bytes]):
        if isinstance(data, bytes):
            data = QtCore.QByteArray(data)
        if isinstance(data, QtCore.QByteArray):
            pixmap = QtGui.QPixmap()
            pixmap.loadFromData(data)
        else:
            pixmap = data
        self.addPixmap(pixmap)

    def get_pixmap(self, size: int) -> QtGui.QPixmap:
        size = core.Size(size, size)
        return self.pixmap(size)


if __name__ == "__main__":
    app = gui.app()
    icon = Icon.for_color("green")



# -*- coding: utf-8 -*-

from qtpy import QtWidgets

from prettyqt import core

QtWidgets.QMouseEventTransition.__bases__ = (core.EventTransition,)


class MouseEventTransition(QtWidgets.QMouseEventTransition):
    pass



# -*- coding: utf-8 -*-

from typing import Optional
from qtpy import QtWidgets, QtGui, QtCore


class ToolTip(QtWidgets.QToolTip):
    @classmethod
    def show_text(
        cls,
        position: Optional[QtCore.QPoint] = None,
        text: str = "",
        linebreak_px: int = 400,
    ):
        if position is None:
            cursor = QtGui.QCursor()
            position = cursor.pos()
        cls.showText(position, f'<div style="max-width: {linebreak_px}px">{text}</div>')


if __name__ == "__main__":
    from prettyqt import widgets

    app = widgets.app()
    ToolTip.show_text(text="test")
    app.main_loop()



# -*- coding: utf-8 -*-

from typing import Any, Callable, Optional, Union, List, Iterator

from qtpy import QtWidgets

from prettyqt import core, gui, widgets


QtWidgets.QMenu.__bases__ = (widgets.Widget,)


class Menu(QtWidgets.QMenu):
    def __init__(
        self,
        title: str = "",
        icon: gui.icon.IconType = None,
        parent: Optional[QtWidgets.QWidget] = None,
    ):
        super().__init__(title, parent=parent)
        self.set_icon(icon)
        self.setToolTipsVisible(True)

    def __iter__(self) -> Iterator[QtWidgets.QAction]:
        return iter(self.actions())

    def __len__(self) -> int:
        return len(self.actions())

    def __add__(self, other):
        if isinstance(other, QtWidgets.QAction):
            self.add(other)
            return self
        raise TypeError("Invalid Type")

    def __getitem__(self, item: str) -> QtWidgets.QAction:
        for action in self.actions():
            if action.objectName() == item:
                return action
        raise KeyError(f"Action {item} not in menu")

    def add(self, *item):
        for i in item:
            self.add_action(i)

    def set_icon(self, icon: gui.icon.IconType):
        """Set the icon for the menu.

        Args:
            icon: icon to use
        """
        icon = gui.icon.get_icon(icon)
        self.setIcon(icon)

    def add_separator(self, text: Optional[str] = None) -> widgets.WidgetAction:
        """Adds a separator showing an optional label.

        Args:
            text: Text to show on separator

        Returns:
            Separator action
        """
        separator = widgets.WidgetAction(parent=self)
        if text is None:
            separator.setSeparator(True)
        else:
            label = widgets.Label(text)
            label.setMinimumWidth(self.minimumWidth())
            with label.edit_stylesheet() as ss:
                ss.background.setValue("lightgrey")
            label.set_alignment(horizontal="center")
            separator.setDefaultWidget(label)
            separator.setEnabled(False)
        self.add(separator)
        return separator

    def add_action(
        self,
        label: Union[str, widgets.Action],
        callback: Callable = None,
        icon: Optional[Any] = None,
        checkable: bool = False,
        checked: bool = False,
        shortcut: Optional[str] = None,
        status_tip: Optional[str] = None,
    ) -> widgets.Action:
        """Add an action to the menu.

        Args:
            label: Label for button
            callback: gets called when action is triggered
            icon: icon for button
            checkable: as checkbox button
            checked: if checkable, turn on by default
            shortcut: Shortcut for action
            status_tip: Status tip to be shown in status bar

        Returns:
            Action added to menu
        """
        if isinstance(label, str):
            action = widgets.Action(text=label, parent=self)
            if callback:
                action.triggered.connect(callback)
            action.set_icon(icon)
            action.set_shortcut(shortcut)
            if checkable:
                action.setCheckable(True)
                action.setChecked(checked)
            if status_tip is not None:
                action.setStatusTip(status_tip)
        else:
            action = label
            action.setParent(self)
        self.addAction(action)
        return action

    def add_actions(self, actions: List[QtWidgets.QAction]):
        self.addActions(actions)

    def add_menu(self, menu: QtWidgets.QMenu) -> QtWidgets.QAction:
        return self.addAction(menu.menuAction())


if __name__ == "__main__":
    app = widgets.app()
    menu = Menu("1")
    action = widgets.Action(text="test")
    menu.addAction(action)
    menu.show()
    menu.exec_(core.Point(200, 200))



from qtpy import QtWidgets

from prettyqt import widgets


QtWidgets.QStyleOptionSizeGrip.__bases__ = (widgets.StyleOptionComplex,)


class StyleOptionSizeGrip(QtWidgets.QStyleOptionSizeGrip):
    pass



# -*- coding: utf-8 -*-

from qtpy import QtGui, QtCore

from prettyqt import core
from prettyqt.utils import bidict, InvalidParamError

PERFORMANCE_HINT = bidict(
    moderate=QtGui.QStaticText.ModerateCaching,
    aggressive=QtGui.QStaticText.AggressiveCaching,
)

TEXT_FORMAT = bidict(
    rich=QtCore.Qt.RichText, plain=QtCore.Qt.PlainText, auto=QtCore.Qt.AutoText
)

if core.VersionNumber.get_qt_version() >= (5, 14, 0):
    TEXT_FORMAT["markdown"] = QtCore.Qt.MarkdownText


class StaticText(QtGui.QStaticText):
    def __repr__(self):
        return f"StaticText({self.text()!r})"

    def __str__(self):
        return self.text()

    def get_size(self) -> core.Size:
        return core.Size(self.size())

    def set_text_format(self, text_format: str):
        """Set the text format.

        Allowed values are "rich", "plain", "auto", "markdown"

        Args:
            text_format: text format to use

        Raises:
            InvalidParamError: text format does not exist
        """
        if text_format not in TEXT_FORMAT:
            raise InvalidParamError(text_format, TEXT_FORMAT)
        self.setTextFormat(TEXT_FORMAT[text_format])

    def get_text_format(self) -> str:
        """Return current text format.

        Possible values: "rich", "plain", "auto", "markdown"

        Returns:
            text format
        """
        return TEXT_FORMAT.inv[self.textFormat()]

    def set_performance_hint(self, hint: str):
        """Set the performance hint.

        Allowed values are "moderate", "aggressive"

        Args:
            hint: performance hint to use

        Raises:
            InvalidParamError: performance hint does not exist
        """
        if hint not in PERFORMANCE_HINT:
            raise InvalidParamError(hint, PERFORMANCE_HINT)
        self.setPerformanceHint(PERFORMANCE_HINT[hint])

    def get_performance_hint(self) -> str:
        """Return current performance hint.

        Possible values: "moderate", "aggressive"

        Returns:
            performance hint
        """
        return PERFORMANCE_HINT.inv[self.performanceHint()]


if __name__ == "__main__":
    text = StaticText()



# -*- coding: utf-8 -*-

from qtpy import QtGui

from prettyqt import gui
from prettyqt.utils import bidict, InvalidParamError


POSITIONS = bidict(
    in_flow=QtGui.QTextFrameFormat.InFlow,
    flow_left=QtGui.QTextFrameFormat.FloatLeft,
    flow_right=QtGui.QTextFrameFormat.FloatRight,
)

BORDER_STYLES = bidict(
    none=QtGui.QTextFrameFormat.BorderStyle_None,
    dotted=QtGui.QTextFrameFormat.BorderStyle_Dotted,
    dashed=QtGui.QTextFrameFormat.BorderStyle_Dashed,
    solid=QtGui.QTextFrameFormat.BorderStyle_Solid,
    double=QtGui.QTextFrameFormat.BorderStyle_Double,
    dot_dash=QtGui.QTextFrameFormat.BorderStyle_DotDash,
    dot_dot_dash=QtGui.QTextFrameFormat.BorderStyle_DotDotDash,
    groove=QtGui.QTextFrameFormat.BorderStyle_Groove,
    ridge=QtGui.QTextFrameFormat.BorderStyle_Ridge,
    inset=QtGui.QTextFrameFormat.BorderStyle_Inset,
    outset=QtGui.QTextFrameFormat.BorderStyle_Outset,
)


QtGui.QTextFrameFormat.__bases__ = (gui.TextFormat,)


class TextFrameFormat(QtGui.QTextFrameFormat):
    def get_height(self) -> gui.TextLength:
        length = self.height()
        return gui.TextLength(length.type(), length.rawValue())

    def get_width(self) -> gui.TextLength:
        length = self.width()
        return gui.TextLength(length.type(), length.rawValue())

    def get_border_brush(self) -> gui.Brush:
        return gui.Brush(self.borderBrush())

    def set_border_style(self, style: str):
        """Set border style.

        Valid values: "none", "dotted", "dashed", "solid", "double", "dot_dash",
                      "dot_dot_dash", "groove", "ridge", "inset", "outset"

        Args:
            style: border style

        Raises:
            InvalidParamError: border style does not exist
        """
        if style not in BORDER_STYLES:
            raise InvalidParamError(style, BORDER_STYLES)
        self.setBorderStyle(BORDER_STYLES[style])

    def get_border_style(self) -> str:
        """Get the current border style.

        Possible values: "none", "dotted", "dashed", "solid", "double", "dot_dash",
                             "dot_dot_dash", "groove", "ridge", "inset", "outset"

        Returns:
            border style
        """
        return BORDER_STYLES.inv[self.borderStyle()]

    def set_page_break_policy(self, policy: str):
        """Set page break policy.

        Valid values: "auto", "always_before", "always_after"

        Args:
            policy: page break policy

        Raises:
            InvalidParamError: page break policy does not exist
        """
        if policy not in gui.textformat.PAGE_BREAK_FLAG:
            raise InvalidParamError(policy, gui.textformat.PAGE_BREAK_FLAG)
        self.setPageBreakPolicy(gui.textformat.PAGE_BREAK_FLAG[policy])

    def get_page_break_policy(self) -> str:
        """Get the current page break policy.

        Possible values: "auto", "always_before", "always_after"

        Returns:
            page break policy
        """
        return gui.textformat.PAGE_BREAK_FLAG.inv[self.pageBreakPolicy()]

    def set_position(self, position: str):
        """Set position.

        Valid values: "in_flow", "flow_left", "flow_right"

        Args:
            position: position

        Raises:
            InvalidParamError: position does not exist
        """
        if position not in POSITIONS:
            raise InvalidParamError(position, POSITIONS)
        self.setPosition(POSITIONS[position])

    def get_position(self) -> str:
        """Get the current position.

        Possible values: "in_flow", "flow_left", "flow_right"

        Returns:
            position
        """
        return POSITIONS.inv[self.position()]



# -*- coding: utf-8 -*-

from typing import Callable, Iterator, Optional, Union, Dict, Tuple

from qtpy import QtCore

from prettyqt import core
from prettyqt.utils import bidict


FLAGS = bidict(
    none=QtCore.QRegularExpression.NoPatternOption,
    ignorecase=QtCore.QRegularExpression.CaseInsensitiveOption,
    dotall=QtCore.QRegularExpression.DotMatchesEverythingOption,
    multiline=QtCore.QRegularExpression.MultilineOption,
    verbose=QtCore.QRegularExpression.ExtendedPatternSyntaxOption,
    inverted_greedyness=QtCore.QRegularExpression.InvertedGreedinessOption,
    dont_capture=QtCore.QRegularExpression.DontCaptureOption,
    unicode=QtCore.QRegularExpression.UseUnicodePropertiesOption,
)

MATCH_TYPES = bidict(
    normal=QtCore.QRegularExpression.NormalMatch,
    prefer_complete=QtCore.QRegularExpression.PartialPreferCompleteMatch,
    prefer_first=QtCore.QRegularExpression.PartialPreferFirstMatch,
    no_match=QtCore.QRegularExpression.NoMatch,
)

MATCH_OPTIONS = bidict(
    none=QtCore.QRegularExpression.NoMatchOption,
    anchored=QtCore.QRegularExpression.AnchoredMatchOption,
)


class RegularExpression(QtCore.QRegularExpression):
    def __init__(
        self,
        pattern: Union[str, QtCore.QRegularExpression] = "",
        flags: int = FLAGS["none"],
    ):
        if isinstance(pattern, QtCore.QRegularExpression):
            super().__init__(pattern)
        else:
            if isinstance(flags, int):
                flags = core.RegularExpression.PatternOptions(flags)
            super().__init__(pattern, flags)

    def __repr__(self):
        return f"RegularExpression({self.pattern()!r})"

    def __reduce__(self):
        return (self.__class__, (self.pattern(), int(self.flags)))

    def globalMatch(self, *args, **kwargs) -> core.RegularExpressionMatchIterator:
        it = super().globalMatch(*args, **kwargs)
        return core.RegularExpressionMatchIterator(it)

    def global_match(
        self,
        text: str,
        offset: int = 0,
        match_type: str = "normal",
        anchored: bool = False,
    ):
        options = MATCH_OPTIONS["anchored"] if anchored else MATCH_OPTIONS["none"]
        return self.globalMatch(text, offset, MATCH_TYPES[match_type], options)

    def match(
        self,
        text: str,
        offset: int = 0,
        match_type: str = "normal",
        anchored: bool = False,
    ) -> core.RegularExpressionMatch:
        if isinstance(match_type, str):
            match_type = MATCH_TYPES[match_type]
        if isinstance(anchored, bool):
            options = MATCH_OPTIONS["anchored"] if anchored else MATCH_OPTIONS["none"]
        else:
            options = anchored
        match = super().match(text, offset, match_type, options)
        return core.RegularExpressionMatch(match)

    def fullmatch(
        self, string: str, pos: int = 0, endpos: Optional[int] = None
    ) -> Optional[core.RegularExpressionMatch]:
        if endpos:
            string = string[:endpos]
        match = super().match(string, pos)
        if match.hasMatch() and len(string) == match.end() - match.start():
            return core.RegularExpressionMatch(match)
        else:
            return None

    def finditer(
        self, string: str, pos: int = 0, endpos: Optional[int] = None
    ) -> Iterator[core.RegularExpressionMatch]:
        for match in self.globalMatch(string[:endpos], offset=pos):
            match.pos = pos
            match.endpos = endpos
            match.string = string
            yield match

    def findall(self, string: str, pos: int = 0, endpos: Optional[int] = None) -> list:
        matches = [m for m in self.globalMatch(string[:endpos], offset=pos)]
        return [m.groups() if len(m.groups()) > 1 else m.group(0) for m in matches]

    def subn(
        self, repl: Union[str, Callable], string: str, count: int = 0
    ) -> Tuple[str, int]:
        result = string
        matches = self.global_match(string)
        matches = list(matches)
        if count > 0:
            matches = matches[:count]
        matches = list(reversed(matches))
        for m in matches:
            to_replace = repl if isinstance(repl, str) else repl(m)
            for j in range(self.groups):
                to_replace = to_replace.replace(fr"\g<{j}>", m.group(j))
            for k in self.groupindex.keys():
                to_replace = to_replace.replace(fr"\g<{k}>", m.group(k))
            result = result[: m.start()] + to_replace + result[m.end() :]
        return (result, min(len(matches), count))

    def sub(self, repl: Union[str, Callable], string: str, count: int = 0) -> str:
        res = self.subn(repl, string, count)
        return res[0]

    def search(self, string: str, pos: int = 0, endpos: Optional[int] = None):
        match = super().match(string[:endpos], pos)
        return match if match.hasMatch() else None

    def split(self, string: str, maxsplit: int = 0):
        raise NotImplementedError()
        # result = list()
        # matches = self.global_match(string)
        # matches = list(matches)
        # if 0 < maxsplit <= len(matches):
        #     remainder = string[matches[maxsplit - 1].end() :]
        #     print(remainder)
        # else:
        #     print(None)
        #     remainder = None
        # if maxsplit > 0:
        #     matches = matches[:maxsplit]
        # prev_match = None
        # m = matches[0]
        # if m.start() == 0:
        #     result.append("")
        # else:
        #     result.append(string[0 : m.start()])
        # for g in m.groups():
        #     result.append(g)
        # prev_match = m
        # for m in matches[1:]:
        #     result.append(string[prev_match.end() : m.start()])
        #     for g in m.groups():
        #         result.append(g)
        #     if m.end() == len(string):
        #         result.append("")
        #     prev_match = m
        # if remainder:
        #     result.append(remainder)
        # return result

    @property
    def groups(self) -> int:
        return self.captureCount()

    @property
    def groupindex(self) -> Dict[str, int]:
        return {k: i for i, k in enumerate(self.namedCaptureGroups()[1:], start=1)}

    @property
    def flags(self):
        return self.patternOptions()


if __name__ == "__main__":
    reg = RegularExpression()
    reg.setPattern("-{1,2}")



# -*- coding: utf-8 -*-

from __future__ import annotations

from typing import Union, Optional

import pathlib

from qtpy import QtCore

from prettyqt.utils import bidict


COMPONENT_FORMATTING_OPTIONS = bidict(
    pretty_decoded=QtCore.QUrl.PrettyDecoded,
    encode_spaces=QtCore.QUrl.EncodeSpaces,
    encode_unicode=QtCore.QUrl.EncodeUnicode,
    encode_delimiters=QtCore.QUrl.EncodeDelimiters,
    encode_reserved=QtCore.QUrl.EncodeReserved,
    decode_reserved=QtCore.QUrl.DecodeReserved,
    fully_encoded=QtCore.QUrl.FullyEncoded,
    fully_decoded=QtCore.QUrl.FullyDecoded,
)


PARSING_MODES = bidict(
    tolerant=QtCore.QUrl.TolerantMode,
    strict=QtCore.QUrl.StrictMode,
    decoded=QtCore.QUrl.DecodedMode,
)

FORMATTING_OPTIONS = bidict(
    none=0,  # QtCore.QUrl.None
    remove_scheme=QtCore.QUrl.RemoveScheme,
    remove_password=QtCore.QUrl.RemovePassword,
    remove_user_info=QtCore.QUrl.RemoveUserInfo,
    remove_port=QtCore.QUrl.RemovePort,
    remove_authority=QtCore.QUrl.RemoveAuthority,
    remove_path=QtCore.QUrl.RemovePath,
    remove_query=QtCore.QUrl.RemoveQuery,
    remove_fragment=QtCore.QUrl.RemoveFragment,
    remove_filename=QtCore.QUrl.RemoveFilename,
    prefer_local_file=QtCore.QUrl.PreferLocalFile,
    strip_trailing_slash=QtCore.QUrl.StripTrailingSlash,
    normalize_path_segments=QtCore.QUrl.NormalizePathSegments,
)


class Url(QtCore.QUrl):
    def __init__(self, path: Union[QtCore.QUrl, str, pathlib.Path] = None):
        if path is None:
            super().__init__()
        else:
            super().__init__(str(path) if not isinstance(path, QtCore.QUrl) else path)
            if isinstance(path, pathlib.Path):
                self.setScheme("file")

    # def __str__(self):
    #     return self.absolutePath()

    def __repr__(self):
        return f"core.Url('{self.toString(self.PreferLocalFile)}')"

    def __str__(self):
        return self.toString(self.PreferLocalFile)

    def serialize_fields(self):
        return dict(path=self.toString())

    def serialize(self):
        return self.serialize_fields()

    def to_string(self, prefer_local: bool = False) -> str:
        flags = 0
        if prefer_local:
            flags |= self.PreferLocalFile
        return self.toString(flags)

    def to_path(self) -> pathlib.Path:
        """Get pathlib object from the URL.

        Returns:
            Path
        """
        return pathlib.Path(str(self))

    def is_local_file(self) -> bool:
        return self.isLocalFile()

    @classmethod
    def from_user_input(cls, url: str, working_dir: Optional[str] = None) -> Url:
        if working_dir is None:
            working_dir = ""
        return cls(cls.fromUserInput(url, working_dir))



# -*- coding: utf-8 -*-

from qtpy import QtWidgets

from prettyqt import widgets


QtWidgets.QStyleOptionTabBarBase.__bases__ = (widgets.StyleOption,)


class StyleOptionTabBarBase(QtWidgets.QStyleOptionTabBarBase):
    pass



# -*- coding: utf-8 -*-

from typing import Optional

from qtpy import QtWidgets

from prettyqt import core, widgets


QtWidgets.QDoubleSpinBox.__bases__ = (widgets.AbstractSpinBox,)


class DoubleSpinBox(QtWidgets.QDoubleSpinBox):

    value_changed = core.Signal(float)

    def __init__(
        self,
        parent: Optional[QtWidgets.QWidget] = None,
        min_value: Optional[float] = None,
        max_value: Optional[float] = None,
        default_value: Optional[float] = None,
    ):
        super().__init__(parent)
        self.valueChanged.connect(self.value_changed)
        self.set_range(min_value, max_value)
        if default_value is not None:
            self.set_value(default_value)

    def serialize_fields(self):
        return dict(
            range=(self.minimum(), self.maximum()),
            prefix=self.prefix(),
            suffix=self.suffix(),
            step_type=self.get_step_type(),
            single_step=self.singleStep(),
            value=self.value(),
            decimals=self.decimals(),
        )

    def __setstate__(self, state):
        self.__init__()
        self.set_range(*state["range"])
        self.setValue(state["value"])
        self.setEnabled(state.get("enabled", True))
        self.setToolTip(state.get("tool_tip", ""))
        self.setStatusTip(state.get("status_tip", ""))
        self.setPrefix(state["prefix"])
        self.setSuffix(state["suffix"])
        self.setDecimals(state["decimals"])
        self.setSingleStep(state["single_step"])
        self.set_step_type(state["step_type"])
        self.set_correction_mode(state["correction_mode"])
        self.set_button_symbols(state["button_symbols"])

    def set_range(self, start, end):
        if start is None:
            start = -float("inf")
        if end is None:
            end = float("inf")
        self.setRange(start, end)


if __name__ == "__main__":
    app = widgets.app()
    widget = DoubleSpinBox()
    widget.show()
    app.main_loop()



# -*- coding: utf-8 -*-

from typing import Optional

from qtpy import QtWidgets

from prettyqt import gui, widgets


class IconLabel(widgets.Widget):
    def __init__(
        self,
        text: Optional[str] = None,
        tooltip: str = "",
        icon: gui.icon.IconType = "mdi.help-circle-outline",
        parent: Optional[QtWidgets.QWidget] = None,
    ):
        super().__init__(parent=parent)
        self.set_layout("horizontal")
        self.label = widgets.Label(text)
        self.label.setMargin(10)
        self.label.set_size_policy(horizontal="minimum")
        self.tooltip = tooltip
        icon = gui.icon.get_icon(icon)
        self.icon = widgets.Label()
        self.icon.setToolTip(tooltip)
        self.icon.set_size_policy(horizontal="minimum")
        pixmap = icon.pixmap(20, 20)
        self.icon.setPixmap(pixmap)
        self.box += self.label
        self.box += self.icon
        self.box.setSpacing(0)
        self.box.addStretch()

    def __getattr__(self, value):
        return self.label.__getattribute__(value)

    def __repr__(self):
        return f"IconLabel({self.text()!r})"


if __name__ == "__main__":
    app = widgets.app()
    widget = IconLabel("test", tooltip="testus")
    widget.show()
    app.main_loop()



# -*- coding: utf-8 -*-

from qtpy import PYQT5, PYSIDE2

if PYQT5:
    from PyQt5 import QtPositioning
elif PYSIDE2:
    from PySide2 import QtPositioning


from prettyqt import positioning


class GeoLocation(QtPositioning.QGeoLocation):
    def get_address(self) -> positioning.GeoAddress:
        return positioning.GeoAddress(self.address())

    def get_coordinate(self) -> positioning.GeoCoordinate:
        return positioning.GeoCoordinate(self.coordinate())

    def get_bounding_box(self) -> positioning.GeoRectangle:
        return positioning.GeoRectangle(self.boundingBox())



# -*- coding: utf-8 -*-

from qtpy import QtCore


class ByteArrayMatcher(QtCore.QByteArrayMatcher):
    def __repr__(self):
        return f"{type(self).__name__}({bytes(self.pattern())!r})"


if __name__ == "__main__":
    matcher = ByteArrayMatcher(b"Test")
    print(repr(matcher))



# -*- coding: utf-8 -*-

from typing import Union
import contextlib

from qtpy import QtCore, QtGui

from prettyqt import core, gui
from prettyqt.utils import bidict, colors, InvalidParamError


COMP_MODES = bidict(
    source_over=QtGui.QPainter.CompositionMode_SourceOver,
    destination_over=QtGui.QPainter.CompositionMode_DestinationOver,
    clear=QtGui.QPainter.CompositionMode_Clear,
    source=QtGui.QPainter.CompositionMode_Source,
    destination=QtGui.QPainter.CompositionMode_Destination,
    source_in=QtGui.QPainter.CompositionMode_SourceIn,
    destination_in=QtGui.QPainter.CompositionMode_DestinationIn,
    source_out=QtGui.QPainter.CompositionMode_SourceOut,
    destination_out=QtGui.QPainter.CompositionMode_DestinationOut,
    source_atop=QtGui.QPainter.CompositionMode_SourceAtop,
    destination_atop=QtGui.QPainter.CompositionMode_DestinationAtop,
)

PATTERNS = bidict(
    solid=QtCore.Qt.SolidPattern,
    none=QtCore.Qt.NoBrush,
    cross=QtCore.Qt.CrossPattern,
    linear_gradient=QtCore.Qt.LinearGradientPattern,
    radial_gradient=QtCore.Qt.RadialGradientPattern,
)

CLIP_OPERATIONS = bidict(
    none=QtCore.Qt.NoClip,
    replace=QtCore.Qt.ReplaceClip,
    intersect=QtCore.Qt.IntersectClip,
)

RENDER_HINTS = bidict(
    antialiasing=QtGui.QPainter.Antialiasing,
    text_antialiasing=QtGui.QPainter.TextAntialiasing,
    smooth_pixmap_transform=QtGui.QPainter.SmoothPixmapTransform,
    high_quality_antialiasing=QtGui.QPainter.HighQualityAntialiasing,
    noncosmetic_default_pen=QtGui.QPainter.NonCosmeticDefaultPen,
    qt4_compatible_painting=QtGui.QPainter.Qt4CompatiblePainting,
    lossless_image_rendering=QtGui.QPainter.LosslessImageRendering,
)

FILL_RULES = bidict(odd_even=QtCore.Qt.OddEvenFill, winding=QtCore.Qt.WindingFill)


class Painter(QtGui.QPainter):
    def __enter__(self):
        return self

    def __exit__(self, typ, value, traceback):
        self.end()

    @contextlib.contextmanager
    def paint_on(self, obj):
        self.begin(obj)
        yield self
        self.end()

    @contextlib.contextmanager
    def backup_state(self):
        self.save()
        yield self
        self.restore()

    def draw_image(
        self,
        target: Union[QtCore.QPoint, QtCore.QPointF, QtCore.QRect, QtCore.QRectF],
        frame_buffer: QtGui.QImage,
    ):
        self.set_composition_mode("source_atop")
        self.drawImage(target, frame_buffer)

    def draw_polygon(self, *args, fill_rule: str = "odd_even"):
        if fill_rule not in FILL_RULES:
            raise InvalidParamError(fill_rule, FILL_RULES)
        self.drawPolygon(*args, fillRule=FILL_RULES[fill_rule])

    def use_antialiasing(self):
        self.setRenderHint(self.Antialiasing, True)

    def fill_rect(
        self, rect: Union[QtCore.QRectF, QtCore.QRect], color, pattern: str = "solid"
    ):
        if pattern not in PATTERNS:
            raise InvalidParamError(pattern, PATTERNS)
        if isinstance(rect, tuple):
            rect = core.Rect(*rect)
        if isinstance(color, str):
            if color not in gui.Color.colorNames():
                raise ValueError("Invalid value for color.")
            color = gui.Color(color)
        if pattern != "solid":
            color = gui.Brush(color, PATTERNS[pattern])
        self.fillRect(rect, color)

    def set_pen(
        self,
        style: str = "solid",
        width: int = 1,
        color: colors.ColorType = "black",
        join_style: str = "bevel",
        cap_style: str = "square",
    ):
        """Set pen to use.

        Args:
            style: pen style to use
            width: pen width
            color: pen color
            join_style: pen join style to use
            cap_style: pen cap style to use
        """
        pen = gui.Pen()
        pen.set_style(style)
        pen.set_cap_style(cap_style)
        pen.set_join_style(join_style)
        pen.setWidth(width)
        pen.set_color(color)
        self.setPen(pen)

    def get_pen(self) -> str:
        """Return current pen.

        Returns:
            current pen
        """
        return gui.Pen(self.pen())

    def set_color(self, color: colors.ColorType):
        color = colors.get_color(color)
        self.setPen(color)

    def set_brush(self, brush: Union[QtGui.QBrush, colors.ColorType]):
        if not isinstance(brush, QtGui.QBrush):
            brush = colors.get_color(brush)
        self.setBrush(brush)

    def set_transparent_background(self, transparent: bool = True):
        mode = QtCore.Qt.TransparentMode if transparent else QtCore.Qt.OpaqueMode
        self.setBackgroundMode(mode)

    def set_composition_mode(self, mode: str):
        """Set the current composition mode.

        Possible values: "source_over", "destination_over", "clear", "source",
                         "destination", "source_in", "destination_in", "source_out",
                         "destination_out", "source_atop", "destination_atop",

        Raises:
            InvalidParamError: composition mode does not exist
        """
        if mode not in COMP_MODES:
            raise InvalidParamError(mode, COMP_MODES)
        self.setCompositionMode(COMP_MODES[mode])

    def get_composition_mode(self) -> str:
        """Get the current composition mode.

        Possible values: "source_over", "destination_over", "clear", "source",
                         "destination", "source_in", "destination_in", "source_out",
                         "destination_out", "source_atop", "destination_atop",

        Returns:
            composition mode
        """
        return COMP_MODES.inv[self.compositionMode()]

    def set_clip_path(self, path: QtGui.QPainterPath, operation: str = "replace"):
        if operation not in CLIP_OPERATIONS:
            raise InvalidParamError(operation, CLIP_OPERATIONS)
        self.setClipPath(path, CLIP_OPERATIONS[operation])

    @contextlib.contextmanager
    def clip_path(self, operation: str = "replace"):
        path = gui.PainterPath()
        yield path
        self.set_clip_path(path, operation)



# -*- coding: utf-8 -*-

from qtpy.QtCharts import QtCharts

from prettyqt import charts


QtCharts.QPercentBarSeries.__bases__ = (charts.AbstractBarSeries,)


class PercentBarSeries(QtCharts.QPercentBarSeries):
    pass



# -*- coding: utf-8 -*-

from typing import Optional, Union

from qtpy import QtCore


class RegularExpressionMatch(QtCore.QRegularExpressionMatch):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.string = None
        self.pos = None
        self.endpos = None

    def __repr__(self):
        return "RegularExpressionMatch()"

    def __getitem__(self, item: Union[int, str]) -> Union[tuple, str]:
        return self.group(item)

    def group(self, *groups: Union[int, str]) -> Union[tuple, str]:
        if len(groups) > 1:
            return tuple(self.captured(i) for i in groups)
        if len(groups) == 0:
            return self.captured(0)
        return self.captured(groups[0])

    def groups(self, default=None) -> tuple:
        if self.lastindex is None:
            return tuple()
        return tuple(
            self.group(i) if i <= self.lastindex else default
            for i in range(self.re.captureCount())
        )

    def groupdict(self, default=None) -> dict:
        if self.lastindex is None:
            return dict()
        groups = [
            self.group(i) if i <= self.lastindex else default
            for i in range(self.re.captureCount())
        ]
        names = self.re.namedCaptureGroups()
        return {names[i]: groups[i] for i in range(self.re.captureCount())}

    def start(self, group: int = 0) -> int:
        return self.capturedStart(group)

    def end(self, group: int = 0) -> int:
        return self.capturedEnd(group)

    def span(self, group: int = 0) -> tuple:
        return (self.capturedStart(group), self.capturedEnd(group))

    @property
    def lastindex(self) -> Optional[int]:
        idx = self.lastCapturedIndex()
        return None if idx == -1 else idx

    @property
    def lastgroup(self) -> Optional[str]:
        if self.lastCapturedIndex() == -1:
            return None
        return self.re.namedCaptureGroups()[self.lastCapturedIndex()]

    @property
    def re(self) -> QtCore.QRegularExpression:
        return self.regularExpression()

    @property
    def partial(self) -> bool:
        return self.hasPartialMatch()


if __name__ == "__main__":
    reg = RegularExpressionMatch()



# -*- coding: utf-8 -*-

from typing import Mapping, Union, Iterable, Dict, Optional, Iterator, Tuple

from qtpy import QtWidgets

from prettyqt import core, widgets
from prettyqt.utils import autoslot


class FlagSelectionWidget(widgets.GroupBox):
    value_changed = core.Signal(int)

    def __init__(
        self,
        label: str = "",
        layout: str = "vertical",
        parent: Optional[QtWidgets.QWidget] = None,
    ):
        super().__init__(title=label, parent=parent)
        self.box = widgets.BoxLayout(layout)
        self.buttons: Dict[widgets.CheckBox, int] = dict()
        self.set_layout(self.box)

    def __iter__(self) -> Iterator[Tuple[widgets.CheckBox, int]]:
        return iter(self.buttons.items())

    def add_items(self, items: Union[Iterable, Mapping]):
        if isinstance(items, Mapping):
            for k, v in items.items():
                self.add(v, k)
        else:
            for i in items:
                if isinstance(i, Iterable):
                    self.add(*i)
                else:
                    raise TypeError("Invalid item type")

    def add(self, title: str, flag: int):
        checkbox = widgets.CheckBox(title)
        checkbox.toggled.connect(self.update_choice)
        self.buttons[checkbox] = flag
        self.box += checkbox

    def current_choice(self) -> int:
        ret_val = 0
        for btn, flag in self.buttons.items():
            if btn.isChecked():
                ret_val |= flag
        return int(ret_val)

    @autoslot.AutoSlot
    def update_choice(self, checked: bool):
        choice = self.current_choice()
        self.value_changed.emit(choice)

    def set_value(self, value: int):
        value = int(value)
        for btn, flag in self.buttons.items():
            btn.setChecked(bool(value & flag))

    def get_value(self):
        return self.current_choice()


if __name__ == "__main__":
    import re

    app = widgets.app()
    widget = FlagSelectionWidget()
    items = {re.MULTILINE: "MultiLine", re.IGNORECASE: "Ignore case"}
    widget.add_items(items)
    widget.show()
    app.main_loop()
    print(widget.get_value())



# -*- coding: utf-8 -*-

from qtpy import QtCore

from prettyqt.utils import bidict


TYPES = bidict(
    system_time=QtCore.QElapsedTimer.SystemTime,
    monotonic_clock=QtCore.QElapsedTimer.MonotonicClock,
    tick_counter=QtCore.QElapsedTimer.TickCounter,
    mach_absolute_time=QtCore.QElapsedTimer.MachAbsoluteTime,
    performance_counter=QtCore.QElapsedTimer.PerformanceCounter,
)


class ElapsedTimer(QtCore.QElapsedTimer):
    def __bool__(self):
        return self.isValid()

    def get_clock_type(self) -> str:
        """Return current clock type.

        Possible values: "system_time", "monotonic_clock", "tick_counter",
                         "mach_absolute_time", "performance_counter"

        Returns:
            clock type
        """
        return TYPES.inv[self.clockType()]



# -*- coding: utf-8 -*-

"""quick module.

contains QtQuick-based classes
"""

from .quickview import QuickView
from .quickitem import QuickItem
from .quickpainteditem import QuickPaintedItem
from .quickwindow import QuickWindow


__all__ = ["QuickView", "QuickItem", "QuickWindow", "QuickPaintedItem"]



"""Module that defines AttributeModel."""

import logging
import inspect
import pprint
from typing import Callable

from qtpy import QtGui
from prettyqt.utils import columnitem
from prettyqt import gui, constants
from prettyqt.objbrowser import treeitem


logger = logging.getLogger(__name__)


_PRETTY_PRINTER = pprint.PrettyPrinter(indent=4)

_ALL_PREDICATES = (
    inspect.isgenerator,
    inspect.istraceback,
    inspect.isframe,
    inspect.iscode,
    inspect.isabstract,
    inspect.isgetsetdescriptor,
    inspect.ismemberdescriptor,
)

CALLABLE_COLOR = QtGui.QBrush(gui.Color("mediumblue"))
REGULAR_COLOR = QtGui.QBrush(gui.Color("black"))

REGULAR_FONT = QtGui.QFont()  # Font for members (non-functions)
SPECIAL_ATTR_FONT = QtGui.QFont()  # Font for __special_attributes__
SPECIAL_ATTR_FONT.setItalic(True)


def get_color(tree_item: treeitem.TreeItem) -> QtGui.QBrush:
    if callable(tree_item.obj):
        return CALLABLE_COLOR
    else:
        return REGULAR_COLOR


def get_font(tree_item: treeitem.TreeItem) -> QtGui.QFont:
    if tree_item.is_special_attribute:
        return SPECIAL_ATTR_FONT
    else:
        return REGULAR_FONT


def get_type(tree_item: treeitem.TreeItem) -> str:
    if inspect.isasyncgen(tree_item.obj):
        return "Async generator"
    if inspect.ismethoddescriptor(tree_item.obj):
        return "Method descriptor"
    if inspect.ismethod(tree_item.obj):
        return "Method"
    elif inspect.isfunction(tree_item.obj):
        return "Function"
    elif inspect.isgeneratorfunction(tree_item.obj):
        return "Generator function"
    elif inspect.isclass(tree_item.obj):
        return "Class"
    elif inspect.ismodule(tree_item.obj):
        return "Module"
    elif inspect.isdatadescriptor(tree_item.obj):
        return "Data descriptor"
    elif inspect.isbuiltin(tree_item.obj):
        return "Builtin"
    return "Attribute"


def safe_tio_call(obj_fn: Callable, tree_item, log_exceptions: bool = False) -> str:
    """Call the obj_fn(tree_item.obj).

    Returns empty string in case of an error.
    """
    tio = tree_item.obj
    try:
        return str(obj_fn(tio))
    except Exception as ex:
        if log_exceptions:
            logger.exception(ex)
        return ""


def safe_data_fn(obj_fn: Callable, log_exceptions: bool = False) -> Callable:
    """Create a function that returns an empty string in case of an exception.

    :param fnobj_fn: function that will be wrapped
    :type obj_fn: object to basestring function
    :returns: function that can be used as AttributeModel data_fn attribute
    :rtype: objbrowser.treeitem.TreeItem to string function
    """

    def data_fn(tree_item: treeitem.TreeItem):
        """Call the obj_fn(tree_item.obj).

        Returns empty string in case of an error
        """
        return safe_tio_call(obj_fn, tree_item, log_exceptions=log_exceptions)

    return data_fn


def tio_predicates(tree_item: treeitem.TreeItem) -> str:
    """Return the inspect module predicates that are true for this object."""
    predicates = [pred.__name__ for pred in _ALL_PREDICATES if pred(tree_item.obj)]
    return ", ".join(predicates)


def tio_is_attribute(tree_item: treeitem.TreeItem) -> str:
    """Return 'True' if the tree item object is an attribute of the parent."""
    if tree_item.is_attribute is None:
        return ""
    else:
        return str(tree_item.is_attribute)


def tio_doc_str(tree_item: treeitem.TreeItem) -> str:
    """Return the doc string of an object."""
    try:
        return tree_item.obj.__doc__
    except AttributeError:
        return "<no doc string found>"


#######################
# Column definitions ##
#######################

ATTR_MODEL_NAME = columnitem.ColumnItem(
    name="Name",
    doc="The name of the object.",
    label=lambda tree_item: tree_item.obj_name if tree_item.obj_name else "<root>",
    col_visible=True,
    width=columnitem.SMALL_COL_WIDTH,
    font=get_font,
    foreground_color=get_color,
)

ATTR_MODEL_DESCRIPTION = columnitem.ColumnItem(
    name="Description",
    doc="Description of the object.",
    label=lambda tree_item: get_type(tree_item),
    col_visible=True,
    width=columnitem.SMALL_COL_WIDTH,
)

ATTR_MODEL_PATH = columnitem.ColumnItem(
    name="Path",
    doc="A path to the data: e.g. var[1]['a'].item",
    label=lambda tree_item: tree_item.obj_path if tree_item.obj_path else "<root>",
    col_visible=True,
    width=columnitem.MEDIUM_COL_WIDTH,
)


ATTR_MODEL_STR = columnitem.ColumnItem(
    name="str",
    doc="""The string representation of the object using the str() function.""",
    label=lambda tree_item: str(tree_item.obj),
    col_visible=False,
    width=columnitem.MEDIUM_COL_WIDTH,
    line_wrap="boundary_or_anywhere",
)

ATTR_MODEL_REPR = columnitem.ColumnItem(
    name="repr",
    doc="The string representation of the object using the repr() function.",
    label=lambda tree_item: repr(tree_item.obj),
    col_visible=True,
    width=columnitem.MEDIUM_COL_WIDTH,
    line_wrap="boundary_or_anywhere",
)

ATTR_MODEL_TYPE = columnitem.ColumnItem(
    name="Type",
    doc="Type of the object determined using the builtin type() function",
    label=lambda tree_item: str(type(tree_item.obj)),
    col_visible=False,
    width=columnitem.MEDIUM_COL_WIDTH,
)

ATTR_MODEL_CLASS = columnitem.ColumnItem(
    name="Type name",
    doc="The name of the class of the object via obj.__class__.__name__",
    label=lambda tree_item: type(tree_item.obj).__name__,
    col_visible=True,
    width=columnitem.MEDIUM_COL_WIDTH,
)

ATTR_MODEL_LENGTH = columnitem.ColumnItem(
    name="Length",
    doc="The length of the object using the len() function",
    # data_fn     = tio_length,
    label=safe_data_fn(len),
    col_visible=False,
    alignment=constants.ALIGN_RIGHT,
    width=columnitem.SMALL_COL_WIDTH,
)

ATTR_MODEL_ID = columnitem.ColumnItem(
    name="id",
    doc="The identifier of the object with calculated using the id() function",
    label=lambda tree_item: f"0x{id(tree_item.obj):X}",
    col_visible=False,
    alignment=constants.ALIGN_RIGHT,
    width=columnitem.SMALL_COL_WIDTH,
)

ATTR_MODEL_IS_ATTRIBUTE = columnitem.ColumnItem(
    name="Attribute",
    doc="""The object is an attribute of the parent (opposed to e.g. a list element).
                     Attributes are displayed in italics in the table.
                  """,
    label=tio_is_attribute,
    col_visible=False,
    width=columnitem.SMALL_COL_WIDTH,
)

ATTR_MODEL_CALLABLE = columnitem.ColumnItem(
    name="Callable",
    doc="""True if the object is callable.
                     Determined with the `callable` built-in function.
                     Callable objects are displayed in blue in the table.
                  """,
    label=None,
    checkstate=lambda tree_item: callable(tree_item.obj),
    col_visible=True,
    width=columnitem.SMALL_COL_WIDTH,
)

ATTR_MODEL_IS_ROUTINE = columnitem.ColumnItem(
    name="Routine",
    doc="""True if the object is a user-defined or built-in function or method.
                     Determined with the inspect.isroutine() method.
                  """,
    label=None,
    checkstate=lambda tree_item: inspect.isroutine(tree_item.obj),
    col_visible=False,
    width=columnitem.SMALL_COL_WIDTH,
)

ATTR_MODEL_IS_BUILTIN = columnitem.ColumnItem(
    name="Builtin",
    doc="""True if the object is a builtin.
                     Determined with the inspect.isbuiltin() method.
                  """,
    label=None,
    checkstate=lambda tree_item: inspect.isbuiltin(tree_item.obj),
    col_visible=False,
    width=columnitem.SMALL_COL_WIDTH,
)

ATTR_MODEL_PRED = columnitem.ColumnItem(
    name="Inspect predicates",
    doc="Predicates from the inspect module",
    label=tio_predicates,
    col_visible=False,
    width=columnitem.MEDIUM_COL_WIDTH,
)

ATTR_MODEL_PRETTY_PRINT = columnitem.ColumnItem(
    name="Pretty print",
    doc="Pretty printed representation of the object using the pprint module.",
    label=lambda tree_item: _PRETTY_PRINTER.pformat(tree_item.obj),
    col_visible=False,
    width=columnitem.MEDIUM_COL_WIDTH,
)


ATTR_MODEL_GET_DOC = columnitem.ColumnItem(
    name="inspect.getdoc",
    doc="The object's doc string, leaned up by inspect.getdoc()",
    label=safe_data_fn(inspect.getdoc),
    col_visible=False,
    width=columnitem.MEDIUM_COL_WIDTH,
)

ATTR_MODEL_GET_COMMENTS = columnitem.ColumnItem(
    name="inspect.getcomments",
    doc="Comments above the object's definition. Retrieved using inspect.getcomments()",
    label=lambda tree_item: inspect.getcomments(tree_item.obj),
    col_visible=False,
    width=columnitem.MEDIUM_COL_WIDTH,
)

ATTR_MODEL_GET_MODULE = columnitem.ColumnItem(
    name="inspect.getmodule",
    doc="The object's module. Retrieved using inspect.module",
    label=safe_data_fn(inspect.getmodule),
    col_visible=False,
    width=columnitem.MEDIUM_COL_WIDTH,
)

ATTR_MODEL_GET_FILE = columnitem.ColumnItem(
    name="inspect.getfile",
    doc="The object's file. Retrieved using inspect.getfile",
    label=safe_data_fn(inspect.getfile),
    col_visible=False,
    width=columnitem.MEDIUM_COL_WIDTH,
)

ATTR_MODEL_GET_SOURCE_FILE = columnitem.ColumnItem(
    name="inspect.getsourcefile",  # calls inspect.getfile()
    doc="The object's file. Retrieved using inspect.getsourcefile",
    label=safe_data_fn(inspect.getsourcefile),
    col_visible=False,
    width=columnitem.MEDIUM_COL_WIDTH,
)


ATTR_MODEL_GET_SOURCE = columnitem.ColumnItem(
    name="Inspect.getsource",
    doc="The source code of an object retrieved using inspect.getsource",
    label=safe_data_fn(inspect.getsource),
    col_visible=False,
    width=columnitem.MEDIUM_COL_WIDTH,
)


DEFAULT_ATTR_COLS = [
    ATTR_MODEL_NAME,
    ATTR_MODEL_DESCRIPTION,
    ATTR_MODEL_PATH,
    ATTR_MODEL_STR,
    ATTR_MODEL_REPR,
    ATTR_MODEL_LENGTH,
    ATTR_MODEL_TYPE,
    ATTR_MODEL_CLASS,
    ATTR_MODEL_ID,
    ATTR_MODEL_IS_ATTRIBUTE,
    ATTR_MODEL_CALLABLE,
    ATTR_MODEL_IS_ROUTINE,
    ATTR_MODEL_IS_BUILTIN,
    ATTR_MODEL_PRED,
    ATTR_MODEL_GET_MODULE,
    ATTR_MODEL_GET_FILE,
    ATTR_MODEL_GET_SOURCE_FILE,
]

DEFAULT_ATTR_DETAILS = [
    ATTR_MODEL_PATH,  # to allow for copy/paste
    ATTR_MODEL_STR,  # Too similar to unicode column
    ATTR_MODEL_REPR,
    ATTR_MODEL_PRETTY_PRINT,
    ATTR_MODEL_GET_DOC,
    ATTR_MODEL_GET_COMMENTS,
    # ATTR_MODEL_GET_MODULE, # not used, already in table
    ATTR_MODEL_GET_FILE,
    # ATTR_MODEL_GET_SOURCE_FILE,  # not used, already in table
    ATTR_MODEL_GET_SOURCE,
]

# Sanity check for duplicates
assert len(DEFAULT_ATTR_COLS) == len(set(DEFAULT_ATTR_COLS))
assert len(DEFAULT_ATTR_DETAILS) == len(set(DEFAULT_ATTR_DETAILS))



# # -*- coding: utf-8 -*-

# from qtpy import QtCore

# from prettyqt import core


# QtCore.QRandomGenerator64.__bases__ = (core.RandomGenerator,)


# class RandomGenerator64(QtCore.QRandomGenerator64):
#     pass
